from typing import (
    List,
    Optional,
)


class ServiceManagerInterface:
    def start(self, service: str, instance: Optional[str] = None) -> None:
        """
        service -- name of service to be started
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Start defined service. Raises StartServiceError on failure.
        """
        raise NotImplementedError()

    def stop(self, service: str, instance: Optional[str] = None) -> None:
        """
        service -- name of service to be stopped
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Stop defined service. Raises StopServiceError on failure.
        """
        raise NotImplementedError()

    def enable(self, service: str, instance: Optional[str] = None) -> None:
        """
        service -- name of service to be enabled
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Enable defined service. Raises EnableServiceError on failure.
        """
        raise NotImplementedError()

    def disable(self, service: str, instance: Optional[str] = None) -> None:
        """
        service -- name of service to be disabled
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Disable defined service. Raises DisableServiceError on failure.
        """
        raise NotImplementedError()

    def is_enabled(self, service: str, instance: Optional[str] = None) -> bool:
        """
        service -- name of service to be checked
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Returns True if specified service is enabled, False otherwise.
        """
        raise NotImplementedError()

    def is_running(self, service: str, instance: Optional[str] = None) -> bool:
        """
        service -- name of service to be checked
        instance -- service instance identifier. Available only for system which
            supports multiple service instances (e.g. systemd)

        Returns True if specified service is running (active), False otherwise.
        """
        raise NotImplementedError()

    def is_installed(self, service: str) -> bool:
        """
        service -- name of service to be checked

        Returns True if specified service is installed (managable by init
        system), False otherwise.
        """
        raise NotImplementedError()

    def get_available_services(self) -> List[str]:
        """
        Returns list of service names recognized by init system.
        """
        raise NotImplementedError()

    def is_current_system_supported(self) -> bool:
        """
        Returns True if the instance of this class is able to manage current
        init system.
        """
        raise NotImplementedError()



import sys

from pcs.common.reports import (
    codes,
    ReportItem,
    ReportItemList,
    ReportItemSeverity,
)
from .messages import report_item_msg_from_dto


def warn(message: str, stderr: bool = False) -> None:
    stream = sys.stderr if stderr else sys.stdout
    stream.write(f"Warning: {message}\n")


def error(message: str) -> SystemExit:
    sys.stderr.write(f"Error: {message}\n")
    return SystemExit(1)


def prepare_force_text(report_item: ReportItem) -> str:
    force_text_map = {
        codes.SKIP_OFFLINE_NODES: ", use --skip-offline to override",
    }
    force_code = report_item.severity.force_code
    if force_code:
        return force_text_map.get(force_code, ", use --force to override")
    return ""


def process_library_reports(report_item_list: ReportItemList) -> None:
    if not report_item_list:
        raise error("Errors have occurred, therefore pcs is unable to continue")

    critical_error = False
    for report_item in report_item_list:
        report_dto = report_item.to_dto()
        msg = report_item_msg_from_dto(report_dto.message).message
        severity = report_dto.severity.level

        if severity == ReportItemSeverity.WARNING:
            warn(msg)
            continue

        if severity != ReportItemSeverity.ERROR:
            print(msg)
            continue

        error(
            "{msg}{force}".format(
                msg=msg,
                force=prepare_force_text(report_item),
            )
        )
        critical_error = True

    if critical_error:
        sys.exit(1)



from lxml import etree

from pcs_test.tools.xml import etree_to_str


def error_xml_not_connected():
    return error_xml(
        102,
        "Not connected",
        ["crm_mon: Error: cluster is not available on this node"],
    )


def error_xml(code, message, errors=()):
    dom = etree.Element(
        "pacemaker-result",
        {
            "api-version": "2.3",
            "request": "crm_mon --output-as xml",
        },
    )
    status_el = etree.SubElement(
        dom, "status", {"code": str(code), "message": message}
    )
    if errors:
        errors_el = etree.SubElement(status_el, "errors")
        for error in errors:
            etree.SubElement(errors_el, "error").text = error
    return etree_to_str(dom)


def complete_state(state_xml, resources_xml=None, nodes_xml=None):
    state_dom = etree.fromstring(state_xml)

    if resources_xml:
        resources_dom = etree.fromstring(resources_xml)
        resources_element = state_dom.find("./resources")
        if resources_element is None:
            raise AssertionError(
                "crm_mon.xml template is missing 'resources' element"
            )
        for child in _complete_state_resources(resources_dom).getchildren():
            resources_element.append(child)

    if nodes_xml:
        nodes_dom = etree.fromstring(nodes_xml)
        nodes_element = state_dom.find("./nodes")
        if nodes_element is None:
            raise AssertionError(
                "crm_mon.xml template is missing 'nodes' element"
            )
        for child in _complete_state_nodes(nodes_dom).getchildren():
            nodes_element.append(child)

    # set correct number of nodes and resources into the status
    resources_count = len(
        state_dom.xpath(
            " | ".join(
                [
                    "./resources/bundle",
                    "./resources/clone",
                    "./resources/group",
                    "./resources/resource",
                ]
            )
        )
    )
    nodes_count = len(state_dom.findall("./nodes/node"))
    state_dom.find("./summary/nodes_configured").set("number", str(nodes_count))
    state_dom.find("./summary/resources_configured").set(
        "number", str(resources_count)
    )

    return state_dom


def _complete_state_resources(resources_status):
    for resource in resources_status.iterfind(".//resource"):
        _default_element_attributes(
            resource,
            {
                "active": "true",
                "failed": "false",
                "failure_ignored": "false",
                "managed": "true",
                "nodes_running_on": "1",
                "orphaned": "false",
                "resource_agent": "ocf::pacemaker:Dummy",
                "role": "Started",
            },
        )
    for group in resources_status.xpath(".//group"):
        _default_element_attributes(
            group,
            {
                "disabled": "false",
                "managed": "true",
            },
        )
    for clone in resources_status.xpath(".//clone"):
        _default_element_attributes(
            clone,
            {
                "disabled": "false",
                "failed": "false",
                "failure_ignored": "false",
            },
        )
    for bundle in resources_status.xpath(".//bundle"):
        _default_element_attributes(
            bundle,
            {
                "failed": "false",
                "image": "image:name",
                "type": "docker",
                "unique": "false",
            },
        )
    return resources_status


def _complete_state_nodes(nodes_status):
    for node in nodes_status.iterfind("./node"):
        _default_element_attributes(
            node,
            {
                "expected_up": "true",
                "is_dc": "false",
                "maintenance": "false",
                "online": "true",
                "pending": "false",
                "resources_running": "0",
                "shutdown": "false",
                "standby": "false",
                "standby_onfail": "false",
                "type": "member",
                "unclean": "false",
            },
        )
    return nodes_status


def _default_element_attributes(element, default_attributes):
    for name, value in default_attributes.items():
        if name not in element.attrib:
            element.attrib[name] = value



import logging
from unittest import mock

from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.lib.env import LibraryEnvironment as Env


def get_mocked_env(**kwargs):
    return Env(
        logger=mock.MagicMock(logging.Logger),
        report_processor=MockLibraryReportProcessor(),
        **kwargs,
    )



from unittest import TestCase


class Destroy(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_setup.SetupSuccessMinimal
        pcs_test.tier0.lib.commands.cluster.test_setup.SetupSuccessAddresses
        pcs_test.tier0.lib.commands.cluster.test_setup.Setup2NodeSuccessMinimal
        pcs_test.tier0.lib.commands.cluster.test_setup.SetupWithWait
        pcs_test.tier0.lib.commands.cluster.test_setup.Failures
            .test_cluster_destroy_failure
    """


class DestroyWarnOnFailure(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes.SuccessMinimal
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes
            .FailureClusterDestroy
    """


class GetQuorumStatus(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes.SuccessMinimal
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes.QuorumCheck
        pcs_test.tier0.lib.commands.cluster.test_remove_nodes.FailureQuorumLoss
    """



from lxml.etree import _Element

from pcs.common.types import CibRuleInEffectStatus
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.live import get_rule_in_effect_status
from pcs.lib.xml_tools import etree_to_str


class RuleInEffectEval:
    """
    Base class for evaluating if a rule is in effect. Use one of its descendants
    based on which one suits your needs and environment the best.
    """

    def get_rule_status(self, rule_id: str) -> CibRuleInEffectStatus:
        """
        Figure out if a rule is expired, in effect, not yet in effect

        rule_id -- ID of the rule to be evaluated
        """
        raise NotImplementedError()


class RuleInEffectEvalDummy(RuleInEffectEval):
    """
    Evaluate all rules to UNKNOWN status. This is useful for example when we
    don't want to get slowed down by evaluating the rules.
    """

    def get_rule_status(self, rule_id: str) -> CibRuleInEffectStatus:
        return CibRuleInEffectStatus.UNKNOWN


class RuleInEffectEvalOneByOne(RuleInEffectEval):
    """
    Evaluate a rule by running a pacemaker tool.
    """

    def __init__(self, cib: _Element, runner: CommandRunner):
        """
        cib -- the whole cib containing the rule expressions
        runner -- a class for running external processes
        """
        self._runner = runner
        self._cib = cib
        self._cib_xml = etree_to_str(cib)

    def get_rule_status(self, rule_id: str) -> CibRuleInEffectStatus:
        return get_rule_in_effect_status(self._runner, self._cib_xml, rule_id)


# TODO Implement this once the pacemaker tool is capable of evaluating more
# than one rule per go. If we can eval all rules by running the tool once, it
# would be a significant speedup comparing to running the tool for each rule.
# class RuleInEffectEvalAllAtOnce(RuleInEffectEval):
#     pass



from unittest import mock, TestCase

from pcs_test.tools.assertions import assert_raise_library_error

from pcs import settings
from pcs.common.reports import ReportItemSeverity as Severities
from pcs.common.reports import codes as report_codes
from pcs.lib.external import CommandRunner
import pcs.lib.booth.status as lib


class GetDaemonStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_daemon_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_daemon_status(self.mock_run, "name"))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status", "-c", "name"]
        )

    def test_daemon_not_running(self):
        self.mock_run.run.return_value = ("", "error", 7)
        self.assertEqual("", lib.get_daemon_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_daemon_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_DAEMON_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )


class GetTicketsStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_tickets_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual(
            "output", lib.get_tickets_status(self.mock_run, "name")
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list", "-c", "name"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_tickets_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_TICKET_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list"]
        )


class GetPeersStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_peers_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_peers_status(self.mock_run, "name"))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers", "-c", "name"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_peers_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_PEERS_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers"]
        )



from pcs.common import (
    file as pcs_file,
    file_type_codes,
    reports,
)
from pcs.common.reports.item import ReportItem
from pcs.cli.reports import output
from pcs.cli.file import metadata
from pcs.lib.errors import LibraryError


def middleware_config(config_path, key_path):
    if config_path and not key_path:
        raise output.error(
            "When --booth-conf is specified, "
            "--booth-key must be specified as well"
        )
    if key_path and not config_path:
        raise output.error(
            "When --booth-key is specified, "
            "--booth-conf must be specified as well"
        )
    is_mocked_environment = config_path and key_path

    if is_mocked_environment:
        config_file = pcs_file.RawFile(
            metadata.for_file_type(file_type_codes.BOOTH_CONFIG, config_path)
        )
        key_file = pcs_file.RawFile(
            metadata.for_file_type(file_type_codes.BOOTH_KEY, key_path)
        )

    def create_booth_env():
        try:
            config_data = config_file.read() if config_file.exists() else None
            key_data = key_file.read() if key_file.exists() else None
        # TODO write custom error handling, do not use pcs.lib specific code
        # and LibraryError
        except pcs_file.RawFileError as e:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.FileIoError(
                        e.metadata.file_type_code,
                        e.action,
                        e.reason,
                        file_path=e.metadata.path,
                    )
                )
            ) from e
        return {
            "config_data": config_data,
            "key_data": key_data,
            "key_path": key_path,
        }

    def flush(modified_env):
        if not is_mocked_environment:
            return
        if not modified_env:
            # TODO now this would not happen
            # for more information see comment in
            # pcs.cli.common.lib_wrapper.lib_env_to_cli_env
            raise output.error("Error during library communication")
        try:
            key_file.write(
                modified_env["key_file"]["content"], can_overwrite=True
            )
            config_file.write(
                modified_env["config_file"]["content"], can_overwrite=True
            )
        # TODO write custom error handling, do not use pcs.lib specific code
        # and LibraryError
        except pcs_file.RawFileError as e:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.FileIoError(
                        e.metadata.file_type_code,
                        e.action,
                        e.reason,
                        file_path=e.metadata.path,
                    )
                )
            ) from e

    def apply(next_in_line, env, *args, **kwargs):
        env.booth = create_booth_env() if is_mocked_environment else {}
        result_of_next = next_in_line(env, *args, **kwargs)
        if is_mocked_environment:
            flush(env.booth["modified_env"])
        return result_of_next

    return apply



from pcs.daemon import ruby_pcsd
from pcs.daemon.app.common import BaseHandler


class Sinatra(BaseHandler):
    """
    Sinatra is base class for handlers which calls the Sinatra via wrapper.
    It accept ruby wrapper during initialization. It also provides method for
    transformation result from sinatra to http response.
    """

    def initialize(self, ruby_pcsd_wrapper: ruby_pcsd.Wrapper):
        # pylint: disable=arguments-differ, attribute-defined-outside-init
        self.__ruby_pcsd_wrapper = ruby_pcsd_wrapper

    def send_sinatra_result(self, result: ruby_pcsd.SinatraResult):
        for name, value in result.headers.items():
            self.set_header(name, value)
        self.set_status(result.status)
        self.write(result.body)

    @property
    def ruby_pcsd_wrapper(self):
        return self.__ruby_pcsd_wrapper



from textwrap import dedent
from unittest import mock

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.cib.resource.bundle import (
    GENERIC_CONTAINER_OPTIONS,
    NETWORK_OPTIONS,
    PORT_MAP_OPTIONS,
    STORAGE_MAP_OPTIONS,
)
from pcs.lib.errors import LibraryError

TIMEOUT = 10


class FixturesMixin:
    container_type = None
    bundle_id = None
    image = None

    @property
    def fixture_resources_bundle_simple(self):
        return """
            <resources>
                <bundle id="{bundle_id}">
                    <{container_type} image="{image}" />
                </bundle>
            </resources>
        """.format(
            container_type=self.container_type,
            bundle_id=self.bundle_id,
            image=self.image,
        )


class SetUpMixin:
    initial_resources = "<resources/>"
    initial_cib_filename = "cib-empty.xml"

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            filename=self.initial_cib_filename,
            resources=self.initial_resources,
        )


class UpgradeMixin(FixturesMixin):
    old_version_cib_filename = None

    def test_cib_upgrade(self):
        (
            self.config.runner.cib.load(
                name="load_cib_old_version",
                filename=self.old_version_cib_filename,
                before="runner.cib.load",
            )
            .runner.cib.upgrade(before="runner.cib.load")
            .env.push_cib(resources=self.fixture_resources_bundle_simple)
        )

        self.run_bundle_cmd()

        self.env_assist.assert_reports(
            [fixture.info(report_codes.CIB_UPGRADE_SUCCESSFUL)]
        )


class ParametrizedContainerMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None
    initial_cib_filename = "cib-empty-3.2.xml"

    def test_all_options(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type}
                            image="{image}"
                            network="extra network settings"
                            options="extra options"
                            promoted-max="0"
                            replicas="4"
                            replicas-per-host="2"
                            run-command="/bin/true"
                        />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(
            container_options={
                "image": self.image,
                "network": "extra network settings",
                "options": "extra options",
                "promoted-max": "0",
                "run-command": "/bin/true",
                "replicas": "4",
                "replicas-per-host": "2",
            }
        )

    def test_deprecated_options(self):
        # Setting both deprecated options and their new variants is tested in
        # self.test_options_errors. This shows deprecated options emit warning
        # even when not forced.
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" masters="1" />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            ),
        )
        self.run_bundle_cmd(
            container_options={
                "image": self.image,
                "masters": "1",
            },
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.WARNING,
                    report_codes.DEPRECATED_OPTION,
                    {
                        "option_name": "masters",
                        "option_type": "container",
                        "replaced_by": ["promoted-max"],
                    },
                    None,
                ),
            ]
        )

    def test_invalid_container_options(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                container_options={
                    "replicas-per-host": "0",
                    "replicas": "0",
                    "masters": "-1",
                    "promoted-max": "-2",
                },
                force_options=True,
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.WARNING,
                    report_codes.DEPRECATED_OPTION,
                    {
                        "option_name": "masters",
                        "option_type": "container",
                        "replaced_by": ["promoted-max"],
                    },
                    None,
                ),
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {
                        "option_type": "container",
                        "option_names": [
                            "image",
                        ],
                    },
                    None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="masters",
                    option_value="-1",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="promoted-max",
                    option_value="-2",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                (
                    severities.ERROR,
                    report_codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    {
                        "option_names": [
                            "masters",
                            "promoted-max",
                        ],
                        "option_type": "container",
                    },
                    None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="replicas",
                    option_value="0",
                    allowed_values="a positive integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="replicas-per-host",
                    option_value="0",
                    allowed_values="a positive integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_empty_image(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                container_options={
                    "image": "",
                },
                force_options=True,
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="image",
                    option_value="",
                    allowed_values="image name",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_unknow_container_option(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                container_options={
                    "image": self.image,
                    "extra": "option",
                }
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "container",
                        "allowed": sorted(list(GENERIC_CONTAINER_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE,
                ),
            ]
        )

    def test_unknow_container_option_forced(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" extra="option" />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            ),
        )
        self.run_bundle_cmd(
            container_options={
                "image": self.image,
                "extra": "option",
            },
            force_options=True,
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "container",
                        "allowed": sorted(list(GENERIC_CONTAINER_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    None,
                ),
            ]
        )


class NetworkMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None

    def test_all_options(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <network
                            control-port="12345"
                            host-interface="eth0"
                            host-netmask="24"
                            ip-range-start="192.168.100.200"
                        />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            ),
        )
        self.run_bundle_cmd(
            network_options={
                "control-port": "12345",
                "host-interface": "eth0",
                "host-netmask": "24",
                "ip-range-start": "192.168.100.200",
            }
        )

    def test_options_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                network_options={
                    "control-port": "0",
                    "host-netmask": "abc",
                    "extra": "option",
                }
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="control-port",
                    option_value="0",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    force_code=report_codes.FORCE,
                    option_name="host-netmask",
                    option_value="abc",
                    allowed_values="a number of bits of the mask (1..32)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    force_code=report_codes.FORCE,
                    option_names=["extra"],
                    option_type="network",
                    allowed=sorted(NETWORK_OPTIONS),
                    allowed_patterns=[],
                ),
            ]
        )

    def test_options_forced(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <network host-netmask="abc" extra="option" />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            ),
        )
        self.run_bundle_cmd(
            network_options={
                "host-netmask": "abc",
                "extra": "option",
            },
            force_options=True,
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="host-netmask",
                    option_value="abc",
                    allowed_values="a number of bits of the mask (1..32)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                (
                    severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "network",
                        "allowed": sorted(list(NETWORK_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    None,
                ),
            ]
        )


class PortMapMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None

    def test_several_mappings_and_handle_their_ids(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <network>
                            <port-mapping
                                id="{bundle_id}-port-map-1001-1"
                                port="1001"
                            />
                            <port-mapping
                                id="{bundle_id}-port-map-1001"
                                internal-port="2002"
                                port="2000"
                            />
                            <port-mapping
                                id="{bundle_id}-port-map-3000-3300"
                                range="3000-3300"
                            />
                        </network>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(
            port_map=[
                {
                    "port": "1001",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-port-map-1001".format(
                        bundle_id=self.bundle_id
                    ),
                    "port": "2000",
                    "internal-port": "2002",
                },
                {
                    "range": "3000-3300",
                },
            ]
        )

    def test_options_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                port_map=[
                    {},
                    {
                        "id": "not#valid",
                    },
                    {
                        "internal-port": "1000",
                    },
                    {
                        "port": "abc",
                    },
                    {
                        "port": "2000",
                        "range": "3000-4000",
                        "internal-port": "def",
                    },
                ],
                force_options=True,
            )
        )
        self.env_assist.assert_reports(
            [
                # first
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    {
                        "option_type": "port-map",
                        "option_names": ["port", "range"],
                    },
                    None,
                ),
                # second
                (
                    severities.ERROR,
                    report_codes.INVALID_ID_BAD_CHAR,
                    {
                        "invalid_character": "#",
                        "id": "not#valid",
                        "id_description": "port-map id",
                        "is_first_char": False,
                    },
                    None,
                ),
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    {
                        "option_type": "port-map",
                        "option_names": ["port", "range"],
                    },
                    None,
                ),
                # third
                (
                    severities.ERROR,
                    report_codes.PREREQUISITE_OPTION_IS_MISSING,
                    {
                        "option_type": "port-map",
                        "option_name": "internal-port",
                        "prerequisite_type": "port-map",
                        "prerequisite_name": "port",
                    },
                    None,
                ),
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    {
                        "option_type": "port-map",
                        "option_names": ["port", "range"],
                    },
                    None,
                ),
                # fourth
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="port",
                    option_value="abc",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                # fifth
                (
                    severities.ERROR,
                    report_codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    {
                        "option_names": [
                            "port",
                            "range",
                        ],
                        "option_type": "port-map",
                    },
                    None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="internal-port",
                    option_value="def",
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_forceable_options_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                port_map=[
                    {
                        "range": "3000",
                        "extra": "option",
                    },
                ]
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "port-map",
                        "allowed": sorted(list(PORT_MAP_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    force_code=report_codes.FORCE,
                    option_name="range",
                    option_value="3000",
                    allowed_values="port-port",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_forceable_options_errors_forced(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <network>
                            <port-mapping
                                id="{bundle_id}-port-map-3000"
                                extra="option"
                                range="3000"
                            />
                        </network>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )

        self.run_bundle_cmd(
            port_map=[
                {
                    "range": "3000",
                    "extra": "option",
                },
            ],
            force_options=True,
        )

        self.env_assist.assert_reports(
            [
                (
                    severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "port-map",
                        "allowed": sorted(list(PORT_MAP_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    None,
                ),
                fixture.warn(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="range",
                    option_value="3000",
                    allowed_values="port-port",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )


class StorageMapMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None

    def test_several_mappings_and_handle_their_ids(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <storage>
                            <storage-mapping
                                id="{bundle_id}-storage-map-1"
                                source-dir="/tmp/bundle1a"
                                target-dir="/tmp/bundle1b"
                            />
                            <storage-mapping
                                id="{bundle_id}-storage-map"
                                options="extra options 1"
                                source-dir="/tmp/bundle2a"
                                target-dir="/tmp/bundle2b"
                            />
                            <storage-mapping
                                id="{bundle_id}-storage-map-3"
                                source-dir-root="/tmp/bundle3a"
                                target-dir="/tmp/bundle3b"
                            />
                            <storage-mapping
                                id="{bundle_id}-storage-map-2"
                                options="extra options 2"
                                source-dir-root="/tmp/bundle4a"
                                target-dir="/tmp/bundle4b"
                            />
                        </storage>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )

        self.run_bundle_cmd(
            storage_map=[
                {
                    "source-dir": "/tmp/bundle1a",
                    "target-dir": "/tmp/bundle1b",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-storage-map".format(
                        bundle_id=self.bundle_id
                    ),
                    "source-dir": "/tmp/bundle2a",
                    "target-dir": "/tmp/bundle2b",
                    "options": "extra options 1",
                },
                {
                    "source-dir-root": "/tmp/bundle3a",
                    "target-dir": "/tmp/bundle3b",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-storage-map-2".format(
                        bundle_id=self.bundle_id
                    ),
                    "source-dir-root": "/tmp/bundle4a",
                    "target-dir": "/tmp/bundle4b",
                    "options": "extra options 2",
                },
            ]
        )

    def test_options_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                storage_map=[
                    {},
                    {
                        "id": "not#valid",
                        "source-dir": "/tmp/bundle1a",
                        "source-dir-root": "/tmp/bundle1b",
                        "target-dir": "/tmp/bundle1c",
                    },
                ],
                force_options=True,
            )
        )
        self.env_assist.assert_reports(
            [
                # first
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    {
                        "option_type": "storage-map",
                        "option_names": ["source-dir", "source-dir-root"],
                    },
                    None,
                ),
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {
                        "option_type": "storage-map",
                        "option_names": [
                            "target-dir",
                        ],
                    },
                    None,
                ),
                # second
                (
                    severities.ERROR,
                    report_codes.INVALID_ID_BAD_CHAR,
                    {
                        "invalid_character": "#",
                        "id": "not#valid",
                        "id_description": "storage-map id",
                        "is_first_char": False,
                    },
                    None,
                ),
                (
                    severities.ERROR,
                    report_codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    {
                        "option_type": "storage-map",
                        "option_names": ["source-dir", "source-dir-root"],
                    },
                    None,
                ),
            ]
        )

    def test_forceable_options_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(
                storage_map=[
                    {
                        "source-dir": "/tmp/bundle1a",
                        "target-dir": "/tmp/bundle1b",
                        "extra": "option",
                    },
                ]
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "storage-map",
                        "allowed": sorted(list(STORAGE_MAP_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE,
                ),
            ]
        )

    def test_forceable_options_errors_forced(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <storage>
                            <storage-mapping
                                id="{bundle_id}-storage-map"
                                source-dir="/tmp/bundle1a"
                                target-dir="/tmp/bundle1b"
                                extra="option"
                            />
                        </storage>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )

        self.run_bundle_cmd(
            storage_map=[
                {
                    "source-dir": "/tmp/bundle1a",
                    "target-dir": "/tmp/bundle1b",
                    "extra": "option",
                },
            ],
            force_options=True,
        )

        self.env_assist.assert_reports(
            [
                (
                    severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": [
                            "extra",
                        ],
                        "option_type": "storage-map",
                        "allowed": sorted(list(STORAGE_MAP_OPTIONS)),
                        "allowed_patterns": [],
                    },
                    None,
                ),
            ]
        )


class MetaMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None

    def test_success(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <meta_attributes id="{bundle_id}-meta_attributes">
                            <nvpair id="{bundle_id}-meta_attributes-is-managed"
                                name="is-managed" value="false" />
                            <nvpair id="{bundle_id}-meta_attributes-target-role"
                                name="target-role" value="Stopped" />
                        </meta_attributes>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(
            meta_attributes={
                "target-role": "Stopped",
                "is-managed": "false",
            }
        )

    def test_disabled(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <meta_attributes id="{bundle_id}-meta_attributes">
                            <nvpair id="{bundle_id}-meta_attributes-target-role"
                                name="target-role" value="Stopped" />
                        </meta_attributes>
                        <{container_type} image="{image}" />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(ensure_disabled=True)


class AllOptionsMixin(SetUpMixin):
    container_type = None
    bundle_id = None
    image = None

    def test_success(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <{container_type}
                            image="{image}"
                            network="extra network settings"
                            options="extra options"
                            promoted-max="0"
                            replicas="4"
                            replicas-per-host="2"
                            run-command="/bin/true"
                        />
                        <network
                            control-port="12345"
                            host-interface="eth0"
                            host-netmask="24"
                            ip-range-start="192.168.100.200"
                        >
                            <port-mapping
                                id="{bundle_id}-port-map-1001-2"
                                port="1001"
                            />
                            <port-mapping
                                id="{bundle_id}-port-map-1001"
                                internal-port="2002"
                                port="2000"
                            />
                            <port-mapping
                                id="{bundle_id}-port-map-3000-3300"
                                range="3000-3300"
                            />
                        </network>
                        <storage>
                            <storage-mapping
                                id="{bundle_id}-storage-map-1"
                                source-dir="/tmp/bundle1a"
                                target-dir="/tmp/bundle1b"
                            />
                            <storage-mapping
                                id="{bundle_id}-storage-map"
                                options="extra options 1"
                                source-dir="/tmp/bundle2a"
                                target-dir="/tmp/bundle2b"
                            />
                            <storage-mapping
                                id="{bundle_id}-storage-map-2"
                                source-dir-root="/tmp/bundle3a"
                                target-dir="/tmp/bundle3b"
                            />
                            <storage-mapping
                                id="{bundle_id}-port-map-1001-1"
                                options="extra options 2"
                                source-dir-root="/tmp/bundle4a"
                                target-dir="/tmp/bundle4b"
                            />
                        </storage>
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(
            container_options={
                "image": self.image,
                "promoted-max": "0",
                "network": "extra network settings",
                "options": "extra options",
                "run-command": "/bin/true",
                "replicas": "4",
                "replicas-per-host": "2",
            },
            network_options={
                "control-port": "12345",
                "host-interface": "eth0",
                "host-netmask": "24",
                "ip-range-start": "192.168.100.200",
            },
            port_map=[
                {
                    "port": "1001",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-port-map-1001".format(
                        bundle_id=self.bundle_id
                    ),
                    "port": "2000",
                    "internal-port": "2002",
                },
                {
                    "range": "3000-3300",
                },
            ],
            storage_map=[
                {
                    "source-dir": "/tmp/bundle1a",
                    "target-dir": "/tmp/bundle1b",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-storage-map".format(
                        bundle_id=self.bundle_id
                    ),
                    "source-dir": "/tmp/bundle2a",
                    "target-dir": "/tmp/bundle2b",
                    "options": "extra options 1",
                },
                {
                    "source-dir-root": "/tmp/bundle3a",
                    "target-dir": "/tmp/bundle3b",
                },
                {
                    # use an autogenerated id of the previous item
                    "id": "{bundle_id}-port-map-1001-1".format(
                        bundle_id=self.bundle_id
                    ),
                    "source-dir-root": "/tmp/bundle4a",
                    "target-dir": "/tmp/bundle4b",
                    "options": "extra options 2",
                },
            ],
        )


class WaitMixin(FixturesMixin, SetUpMixin):
    initial_resources = "<resources/>"
    bundle_id = None
    image = None

    @property
    def fixture_status_running(self):
        return """
            <resources>
                <bundle id="{bundle_id}" managed="true">
                    <replica id="0">
                        <resource
                            id="{bundle_id}-docker-0"
                            managed="true"
                            role="Started"
                        >
                            <node name="node1" id="1" cached="false"/>
                        </resource>
                    </replica>
                    <replica id="1">
                        <resource
                            id="{bundle_id}-docker-1"
                            managed="true"
                            role="Started"
                        >
                            <node name="node2" id="2" cached="false"/>
                        </resource>
                    </replica>
                </bundle>
            </resources>
        """.format(
            bundle_id=self.bundle_id
        )

    @property
    def fixture_status_not_running(self):
        return """
            <resources>
                <bundle id="{bundle_id}" managed="true">
                    <replica id="0">
                        <resource
                            id="{bundle_id}-docker-0"
                            managed="true"
                            role="Stopped"
                        />
                    </replica>
                    <replica id="1">
                        <resource
                            id="{bundle_id}-docker-1"
                            managed="true"
                            role="Stopped"
                        />
                    </replica>
                </bundle>
            </resources>
        """.format(
            bundle_id=self.bundle_id
        )

    @property
    def fixture_resources_bundle_simple_disabled(self):
        return """
            <resources>
                <bundle id="{bundle_id}">
                    <meta_attributes id="{bundle_id}-meta_attributes">
                        <nvpair id="{bundle_id}-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                    <docker image="{image}" />
                </bundle>
            </resources>
        """.format(
            bundle_id=self.bundle_id, image=self.image
        )

    def test_wait_fail(self):
        wait_error_message = dedent(
            """\
            Pending actions:
                    Action 12: {bundle_id}-node2-stop on node2
            Error performing operation: Timer expired
            """.format(
                bundle_id=self.bundle_id
            )
        ).strip()
        self.config.env.push_cib(
            resources=self.fixture_resources_bundle_simple,
            wait=TIMEOUT,
            exception=LibraryError(
                reports.item.ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(wait_error_message)
                )
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(wait=TIMEOUT),
            [fixture.report_wait_for_idle_timed_out(wait_error_message)],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_ok(self):
        (
            self.config.env.push_cib(
                resources=self.fixture_resources_bundle_simple, wait=TIMEOUT
            ).runner.pcmk.load_state(resources=self.fixture_status_running)
        )
        self.run_bundle_cmd(wait=TIMEOUT)
        self.env_assist.assert_reports(
            [
                fixture.report_resource_running(
                    self.bundle_id, {"Started": ["node1", "node2"]}
                ),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_fail(self):
        (
            self.config.env.push_cib(
                resources=self.fixture_resources_bundle_simple, wait=TIMEOUT
            ).runner.pcmk.load_state(resources=self.fixture_status_not_running)
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(wait=TIMEOUT)
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_not_running(
                    self.bundle_id, severities.ERROR
                ),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_disabled_wait_ok_run_ok(self):
        (
            self.config.env.push_cib(
                resources=self.fixture_resources_bundle_simple_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(resources=self.fixture_status_not_running)
        )
        self.run_bundle_cmd(ensure_disabled=True, wait=TIMEOUT)
        self.env_assist.assert_reports(
            [
                (
                    severities.INFO,
                    report_codes.RESOURCE_DOES_NOT_RUN,
                    {
                        "resource_id": self.bundle_id,
                    },
                    None,
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_disabled_wait_ok_run_fail(self):
        (
            self.config.env.push_cib(
                resources=self.fixture_resources_bundle_simple_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(resources=self.fixture_status_running)
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.run_bundle_cmd(ensure_disabled=True, wait=TIMEOUT)
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_running(
                    self.bundle_id,
                    {"Started": ["node1", "node2"]},
                    severities.ERROR,
                )
            ]
        )



# pylint: disable=too-many-lines
from unittest import TestCase

from pcs_test.tools.misc import outdent

from pcs.lib.corosync import config_parser


class SectionTest(TestCase):
    def test_empty_section(self):
        section = config_parser.Section("mySection")
        self.assertEqual(section.parent, None)
        self.assertEqual(section.get_root(), section)
        self.assertEqual(section.name, "mySection")
        self.assertEqual(section.get_attributes(), [])
        self.assertEqual(section.get_sections(), [])
        self.assertTrue(section.empty)
        self.assertEqual(str(section), "")

    def test_is_section_empty(self):
        section = config_parser.Section("mySection")
        self.assertTrue(section.empty)

        section = config_parser.Section("mySection")
        section.add_attribute("name", "value")
        self.assertFalse(section.empty)

        section = config_parser.Section("mySection")
        section.add_section(config_parser.Section("subSection"))
        self.assertFalse(section.empty)

        section = config_parser.Section("mySection")
        section.add_attribute("name", "value")
        section.add_section(config_parser.Section("subSection"))
        self.assertFalse(section.empty)

    def test_attribute_add(self):
        section = config_parser.Section("mySection")

        section.add_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ],
        )

        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
            ],
        )

        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name2", "value2"],
            ],
        )

    def test_attribute_get(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2a")

        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2a"],
            ],
        )
        self.assertEqual(
            section.get_attributes("name1"),
            [
                ["name1", "value1"],
            ],
        )
        self.assertEqual(
            section.get_attributes("name2"),
            [
                ["name2", "value2"],
                ["name2", "value2a"],
            ],
        )
        self.assertEqual(section.get_attributes("nameX"), [])

    def test_attribute_get_dict(self):
        section = config_parser.Section("mySection")
        self.assertEqual(section.get_attributes_dict(), {})

        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        self.assertEqual(
            section.get_attributes_dict(),
            {
                "name1": "value1",
                "name2": "value2",
                "name3": "value3",
            },
        )

        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name1", "value1A")
        section.add_attribute("name3", "value3A")
        section.add_attribute("name1", "")
        self.assertEqual(
            section.get_attributes_dict(),
            {
                "name1": "",
                "name2": "value2",
                "name3": "value3A",
            },
        )

    def test_attribute_value(self):
        section = config_parser.Section("mySection")
        self.assertEqual(section.get_attribute_value("name"), None)

        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        self.assertEqual(section.get_attribute_value("name", "value"), "value")

        section = config_parser.Section("mySection")
        section.add_attribute("name", "value")
        section.add_attribute("name1", "value1")
        section.add_attribute("name", "valueA")
        section.add_attribute("name1", "value1A")
        self.assertEqual(section.get_attribute_value("name"), "valueA")

    def test_attribute_set(self):
        section = config_parser.Section("mySection")

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ],
        )

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ],
        )

        section.set_attribute("name1", "value1a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1a"],
            ],
        )

        section.set_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1a"],
                ["name2", "value2"],
            ],
        )

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
            ],
        )

        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2"],
            ],
        )
        section.set_attribute("name2", "value2a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
                ["name3", "value3"],
            ],
        )

        section.add_attribute("name1", "value1")
        section.add_attribute("name1", "value1")
        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
                ["name3", "value3"],
            ],
        )

    def test_attribute_change(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        attr = section.get_attributes()[1]
        attr[0] = "name2a"
        attr[1] = "value2a"
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2a", "value2a"],
                ["name3", "value3"],
                ["name2", "value2"],
            ],
        )

    def test_attribute_del(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        section.del_attribute(section.get_attributes()[1])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
            ],
        )

        section.del_attribute(["name3", "value3"])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ],
        )

        section.del_attribute(["name3", "value3"])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ],
        )

    def test_attribute_del_by_name(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        section.del_attributes_by_name("nameX")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2"],
            ],
        )

        section.del_attributes_by_name("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
            ],
        )

        section.add_attribute("name2", "value2")
        section.add_attribute("name2", "value2a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2"],
                ["name2", "value2a"],
            ],
        )
        section.del_attributes_by_name("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2a"],
            ],
        )

        section.add_attribute("name3", "value3a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2a"],
                ["name3", "value3a"],
            ],
        )
        section.del_attributes_by_name("name3")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
            ],
        )

    def test_section_add(self):
        root = config_parser.Section("root")
        child1 = config_parser.Section("child1")
        child1a = config_parser.Section("child1a")
        child2 = config_parser.Section("child2")

        root.add_section(child1)
        child1.add_section(child1a)
        root.add_section(child2)
        self.assertEqual(root.parent, None)
        self.assertEqual(child1.parent.name, "root")
        self.assertEqual(child1a.parent.name, "child1")
        self.assertEqual(child2.parent.name, "root")
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                child1a {
                }
            }

            child2 {
            }
            """
            ),
        )

        child2.add_section(child1a)
        self.assertEqual(child1a.parent.name, "child2")
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
            }

            child2 {
                child1a {
                }
            }
            """
            ),
        )

        self.assertRaises(
            config_parser.CircularParentshipException,
            child1a.add_section,
            child1a,
        )
        self.assertRaises(
            config_parser.CircularParentshipException,
            child1a.add_section,
            child2,
        )
        self.assertRaises(
            config_parser.CircularParentshipException, child1a.add_section, root
        )

    def test_section_get(self):
        root = config_parser.Section("")
        child1 = config_parser.Section("child1")
        child2 = config_parser.Section("child2")
        childa1 = config_parser.Section("childA")
        childa2 = config_parser.Section("childA")
        childa3 = config_parser.Section("childA")
        childa4 = config_parser.Section("childA")
        childb1 = config_parser.Section("childB")
        childb2 = config_parser.Section("childB")
        childa1.add_attribute("id", "1")
        childa2.add_attribute("id", "2")
        childa3.add_attribute("id", "3")
        childa4.add_attribute("id", "4")
        childb1.add_attribute("id", "5")
        childb2.add_attribute("id", "6")
        root.add_section(child1)
        root.add_section(child2)
        child1.add_section(childa1)
        child1.add_section(childa2)
        child1.add_section(childb1)
        child2.add_section(childa3)
        child2.add_section(childb2)
        child2.add_section(childa4)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                childA {
                    id: 1
                }

                childA {
                    id: 2
                }

                childB {
                    id: 5
                }
            }

            child2 {
                childA {
                    id: 3
                }

                childB {
                    id: 6
                }

                childA {
                    id: 4
                }
            }
            """
            ),
        )

        self.assertEqual(
            "---\n".join([str(x) for x in root.get_sections()]),
            outdent(
                """\
                child1 {
                    childA {
                        id: 1
                    }

                    childA {
                        id: 2
                    }

                    childB {
                        id: 5
                    }
                }
                ---
                child2 {
                    childA {
                        id: 3
                    }

                    childB {
                        id: 6
                    }

                    childA {
                        id: 4
                    }
                }
                """
            ),
        )

        self.assertEqual(
            "---\n".join([str(x) for x in root.get_sections("child1")]),
            outdent(
                """\
                child1 {
                    childA {
                        id: 1
                    }

                    childA {
                        id: 2
                    }

                    childB {
                        id: 5
                    }
                }
                """
            ),
        )

        self.assertEqual(
            "---\n".join([str(x) for x in child1.get_sections("childA")]),
            outdent(
                """\
                childA {
                    id: 1
                }
                ---
                childA {
                    id: 2
                }
                """
            ),
        )

        self.assertEqual(
            "---\n".join([str(x) for x in child1.get_sections("child2")]), ""
        )

    def test_section_del(self):
        root = config_parser.Section("")
        child1 = config_parser.Section("child1")
        child2 = config_parser.Section("child2")
        childa1 = config_parser.Section("childA")
        childa2 = config_parser.Section("childA")
        childa3 = config_parser.Section("childA")
        childa4 = config_parser.Section("childA")
        childb1 = config_parser.Section("childB")
        childb2 = config_parser.Section("childB")
        childa1.add_attribute("id", "1")
        childa2.add_attribute("id", "2")
        childa3.add_attribute("id", "3")
        childa4.add_attribute("id", "4")
        childb1.add_attribute("id", "5")
        childb2.add_attribute("id", "6")
        root.add_section(child1)
        root.add_section(child2)
        child1.add_section(childa1)
        child1.add_section(childa2)
        child1.add_section(childb1)
        child2.add_section(childa3)
        child2.add_section(childb2)
        child2.add_section(childa4)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                childA {
                    id: 1
                }

                childA {
                    id: 2
                }

                childB {
                    id: 5
                }
            }

            child2 {
                childA {
                    id: 3
                }

                childB {
                    id: 6
                }

                childA {
                    id: 4
                }
            }
            """
            ),
        )

        child2.del_section(childb2)
        self.assertEqual(childb2.parent, None)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                childA {
                    id: 1
                }

                childA {
                    id: 2
                }

                childB {
                    id: 5
                }
            }

            child2 {
                childA {
                    id: 3
                }

                childA {
                    id: 4
                }
            }
            """
            ),
        )

        root.del_section(child2)
        self.assertEqual(child2.parent, None)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                childA {
                    id: 1
                }

                childA {
                    id: 2
                }

                childB {
                    id: 5
                }
            }
            """
            ),
        )

        self.assertRaises(ValueError, root.del_section, child2)

        self.assertEqual(childa1.parent.name, "child1")
        self.assertRaises(ValueError, child2.del_section, childa1)
        self.assertEqual(childa1.parent.name, "child1")

        child1.del_section(childb1)
        self.assertEqual(childb1.parent, None)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
                childA {
                    id: 1
                }

                childA {
                    id: 2
                }
            }
            """
            ),
        )

        child1.del_section(childa1)
        self.assertEqual(childa1.parent, None)
        child1.del_section(childa2)
        self.assertEqual(childa2.parent, None)
        self.assertEqual(
            str(root),
            outdent(
                """\
            child1 {
            }
            """
            ),
        )

        root.del_section(child1)
        self.assertEqual(child1.parent, None)
        self.assertEqual(str(root), "")

    def test_get_root(self):
        root = config_parser.Section("root")
        child1 = config_parser.Section("child1")
        child1a = config_parser.Section("child1a")
        root.add_section(child1)
        child1.add_section(child1a)

        self.assertEqual(root.get_root().name, "root")
        self.assertEqual(child1.get_root().name, "root")
        self.assertEqual(child1a.get_root().name, "root")

    def test_str(self):
        root = config_parser.Section("root")
        self.assertEqual(str(root), "")

        root.add_attribute("name1", "value1")
        self.assertEqual(str(root), "name1: value1\n")

        root.add_attribute("name2", "value2")
        root.add_attribute("name2", "value2a")
        root.add_attribute("name3", "value3")
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3
            """
            ),
        )

        child1 = config_parser.Section("child1")
        root.add_section(child1)
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3

            child1 {
            }
            """
            ),
        )

        child1.add_attribute("name1.1", "value1.1")
        child1.add_attribute("name1.2", "value1.2")
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3

            child1 {
                name1.1: value1.1
                name1.2: value1.2
            }
            """
            ),
        )

        child2 = config_parser.Section("child2")
        child2.add_attribute("name2.1", "value2.1")
        root.add_section(child2)
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3

            child1 {
                name1.1: value1.1
                name1.2: value1.2
            }

            child2 {
                name2.1: value2.1
            }
            """
            ),
        )

        child2a = config_parser.Section("child2a")
        child2a.add_attribute("name2.a.1", "value2.a.1")
        child2.add_section(child2a)
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3

            child1 {
                name1.1: value1.1
                name1.2: value1.2
            }

            child2 {
                name2.1: value2.1

                child2a {
                    name2.a.1: value2.a.1
                }
            }
            """
            ),
        )

        child3 = config_parser.Section("child3")
        root.add_section(child3)
        child3.add_section(config_parser.Section("child3a"))
        child3.add_section(config_parser.Section("child3b"))
        self.assertEqual(
            str(root),
            outdent(
                """\
            name1: value1
            name2: value2
            name2: value2a
            name3: value3

            child1 {
                name1.1: value1.1
                name1.2: value1.2
            }

            child2 {
                name2.1: value2.1

                child2a {
                    name2.a.1: value2.a.1
                }
            }

            child3 {
                child3a {
                }

                child3b {
                }
            }
            """
            ),
        )


class ParserTest(TestCase):
    # pylint: disable=too-many-public-methods
    def test_empty(self):
        self.assertEqual(
            str(config_parser.Parser.parse("".encode("utf-8"))), ""
        )

    def test_attributes_one_attribute(self):
        string = outdent(
            """\
            name:value\
            """
        )
        parsed = outdent(
            """\
            name: value
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_attributes_two_attributes_same_name(self):
        string = outdent(
            """\
            name:value
            name:value
            """
        )
        parsed = outdent(
            """\
            name: value
            name: value
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_attributes_more_attributes_whitespace(self):
        # pylint: disable=trailing-whitespace
        string = outdent(
            """\
              name1:value1  
            name2  :value2
            name3:  value3
              name4  :  value4  
            """
        )
        parsed = outdent(
            """\
            name1: value1
            name2: value2
            name3: value3
            name4: value4
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_attributes_colon_in_value(self):
        string = outdent(
            """\
            name:foo:value
            """
        )
        parsed = outdent(
            """\
            name: foo:value
            """
        )
        root = config_parser.Parser.parse(string.encode("utf-8"))
        self.assertEqual(root.get_attributes(), [["name", "foo:value"]])
        self.assertEqual(str(root), parsed)

    def test_attributes_empty_value(self):
        # pylint: disable=trailing-whitespace
        string = outdent(
            """\
            name :  
            """
        )
        parsed = outdent(
            """\
            name: 
            """
        )
        root = config_parser.Parser.parse(string.encode("utf-8"))
        self.assertEqual(root.get_attributes(), [["name", ""]])
        self.assertEqual(str(root), parsed)

    def test_sections_empty_section(self):
        string = outdent(
            """\
            section1 {
            }\
            """
        )
        parsed = outdent(
            """\
            section1 {
            }
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_sections_empty_section_in_section_whitespace(self):
        # pylint: disable=trailing-whitespace
        string = outdent(
            """\
            section1 {
                section1a   {
              }
              section1b        {       
                 }    
            }
            """
        )
        parsed = outdent(
            """\
            section1 {
                section1a {
                }

                section1b {
                }
            }
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_sections_no_name_before_opening(self):
        string = outdent(
            """\
            section1 {
                {
                }
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.MissingSectionNameBeforeOpeningBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_junk_after_opening(self):
        string = outdent(
            """\
            section1 {
                section1a {junk
                }
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.ExtraCharactersAfterOpeningBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_comment_junk_after_opening(self):
        string = outdent(
            """\
            section1 {
                section1a { #junk
                }
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.ExtraCharactersAfterOpeningBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_junk_before_closing(self):
        string = outdent(
            """\
            section1 {
                section1a {
                junk}
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.ExtraCharactersBeforeOrAfterClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_junk_after_closing(self):
        string = outdent(
            """\
            section1 {
                section1a {
                }junk
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.ExtraCharactersBeforeOrAfterClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_comment_junk_after_closing(self):
        string = outdent(
            """\
            section1 {
                section1a {
                } #junk
            }
            section2 {
               section2a {
               }
               section2b {
               }
            }
            """
        )
        self.assertRaises(
            config_parser.ExtraCharactersBeforeOrAfterClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_unexpected_closing_brace(self):
        string = outdent(
            """\
            }
            """
        )
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_unexpected_closing_brace_inner_section(self):
        string = outdent(
            """\
            section1 {
                section1a {
                }

                section1b {
                }
            }
            }
            """
        )
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_missing_closing_brace(self):
        string = outdent(
            """\
            section1 {
            """
        )
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_sections_missing_closing_brace_inner_section(self):
        string = outdent(
            """\
            section1 {
                section1a {

                section1b {
                }
            }
            """
        )
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_junk_line(self):
        string = outdent(
            """\
            name1: value1
            section1 {
                section1a {
                    name1a: value1a
                }
            junk line
                section1b {
                }
            }
            """
        )
        self.assertRaises(
            config_parser.LineIsNotSectionNorKeyValueException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_comments_attributes(self):
        string = outdent(
            """\
            # junk1
            name1: value1
              #junk2
            name2: value2#junk3
            name3: value3 #junk4
            name4 # junk5: value4
            #junk6 name5: value5
            #junk7
            """
        )
        parsed = outdent(
            """\
            name1: value1
            name2: value2#junk3
            name3: value3 #junk4
            name4 # junk5: value4
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_comments_sections_closing_brace(self):
        string = outdent(
            """\
            section {
            #}
            """
        )
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_comments_sections_opening_brace(self):
        string = outdent(
            """\
            #section {
            }
            """
        )
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.Parser.parse,
            string.encode("utf-8"),
        )

    def test_full_1(self):
        # pylint: disable=line-too-long
        string = outdent(
            """\
            # Please read the corosync.conf.5 manual page
            totem {
            	version: 2

            	# crypto_cipher and crypto_hash: Used for mutual node authentication.
            	# If you choose to enable this, then do remember to create a shared
            	# secret with "corosync-keygen".
            	# enabling crypto_cipher, requires also enabling of crypto_hash.
            	crypto_cipher: none
            	crypto_hash: none

            	# interface: define at least one interface to communicate
            	# over. If you define more than one interface stanza, you must
            	# also set rrp_mode.
            	interface {
                # Rings must be consecutively numbered, starting at 0.
            		ringnumber: 0
            		# This is normally the *network* address of the
            		# interface to bind to. This ensures that you can use
            		# identical instances of this configuration file
            		# across all your cluster nodes, without having to
            		# modify this option.
            		bindnetaddr: 192.168.1.0
            		# However, if you have multiple physical network
            		# interfaces configured for the same subnet, then the
            		# network address alone is not sufficient to identify
            		# the interface Corosync should bind to. In that case,
            		# configure the *host* address of the interface
            		# instead:
            		# bindnetaddr: 192.168.1.1
            		# When selecting a multicast address, consider RFC
            		# 2365 (which, among other things, specifies that
            		# 239.255.x.x addresses are left to the discretion of
            		# the network administrator). Do not reuse multicast
            		# addresses across multiple Corosync clusters sharing
            		# the same network.
            		mcastaddr: 239.255.1.1
            		# Corosync uses the port you specify here for UDP
            		# messaging, and also the immediately preceding
            		# port. Thus if you set this to 5405, Corosync sends
            		# messages over UDP ports 5405 and 5404.
            		mcastport: 5405
            		# Time-to-live for cluster communication packets. The
            		# number of hops (routers) that this ring will allow
            		# itself to pass. Note that multicast routing must be
            		# specifically enabled on most network routers.
            		ttl: 1
            	}
            }

            logging {
            	# Log the source file and line where messages are being
            	# generated. When in doubt, leave off. Potentially useful for
            	# debugging.
            	fileline: off
            	# Log to standard error. When in doubt, set to no. Useful when
            	# running in the foreground (when invoking "corosync -f")
            	to_stderr: no
            	# Log to a log file. When set to "no", the "logfile" option
            	# must not be set.
            	to_logfile: yes
                logfile: /var/log/cluster/corosync.log
            	# Log to the system log daemon. When in doubt, set to yes.
            	to_syslog: yes
            	# Log debug messages (very verbose). When in doubt, leave off.
            	debug: off
            	# Log messages with time stamps. When in doubt, set to on
            	# (unless you are only logging to syslog, where double
            	# timestamps can be annoying).
            	timestamp: on
            	logger_subsys {
            		subsys: QUORUM
            		debug: off
            	}
            }

            quorum {
            	# Enable and configure quorum subsystem (default: off)
            	# see also corosync.conf.5 and votequorum.5
            	#provider: corosync_votequorum
            }
            """
        )
        parsed = outdent(
            """\
            totem {
                version: 2
                crypto_cipher: none
                crypto_hash: none

                interface {
                    ringnumber: 0
                    bindnetaddr: 192.168.1.0
                    mcastaddr: 239.255.1.1
                    mcastport: 5405
                    ttl: 1
                }
            }

            logging {
                fileline: off
                to_stderr: no
                to_logfile: yes
                logfile: /var/log/cluster/corosync.log
                to_syslog: yes
                debug: off
                timestamp: on

                logger_subsys {
                    subsys: QUORUM
                    debug: off
                }
            }

            quorum {
            }
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )

    def test_full_2(self):
        string = outdent(
            """\
            # Please read the corosync.conf.5 manual page
            totem {
            	version: 2

            	crypto_cipher: none
            	crypto_hash: none

            	interface {
            		ringnumber: 0
            		bindnetaddr: 10.16.35.0
            		mcastport: 5405
            		ttl: 1
            	}
            	transport: udpu
            }

            logging {
            	fileline: off
            	to_logfile: yes
            	to_syslog: yes
                logfile: /var/log/cluster/corosync.log
            	debug: off
            	timestamp: on
            	logger_subsys {
            		subsys: QUORUM
            		debug: off
            	}
            }

            nodelist {
            	node {
            		ring0_addr: 10.16.35.101
            		nodeid: 1
            	}

            	node {
            		ring0_addr: 10.16.35.102
            		nodeid: 2
            	}

            	node {
            		ring0_addr: 10.16.35.103
            	}

            	node {
            		ring0_addr: 10.16.35.104
            	}

            	node {
            		ring0_addr: 10.16.35.105
            	}
            }

            quorum {
            	# Enable and configure quorum subsystem (default: off)
            	# see also corosync.conf.5 and votequorum.5
            	#provider: corosync_votequorum
            }
            """
        )
        parsed = outdent(
            """\
            totem {
                version: 2
                crypto_cipher: none
                crypto_hash: none
                transport: udpu

                interface {
                    ringnumber: 0
                    bindnetaddr: 10.16.35.0
                    mcastport: 5405
                    ttl: 1
                }
            }

            logging {
                fileline: off
                to_logfile: yes
                to_syslog: yes
                logfile: /var/log/cluster/corosync.log
                debug: off
                timestamp: on

                logger_subsys {
                    subsys: QUORUM
                    debug: off
                }
            }

            nodelist {
                node {
                    ring0_addr: 10.16.35.101
                    nodeid: 1
                }

                node {
                    ring0_addr: 10.16.35.102
                    nodeid: 2
                }

                node {
                    ring0_addr: 10.16.35.103
                }

                node {
                    ring0_addr: 10.16.35.104
                }

                node {
                    ring0_addr: 10.16.35.105
                }
            }

            quorum {
            }
            """
        )
        self.assertEqual(
            str(config_parser.Parser.parse(string.encode("utf-8"))), parsed
        )


class VerifySection(TestCase):
    def test_empty_section(self):
        section = config_parser.Section("mySection")
        self.assertEqual(config_parser.verify_section(section), ([], [], []))

    def test_all_valid(self):
        text = outdent(
            """\
            name1: value1
            name2: value2

            child1 {
                name1_1: value1.1
                name1_2: value1.2

                child1A {
                    name1A1: value
                }
                child1B {
                    name1B1: value
                    name1B2: value
                }
            }

            child2 {
                child2 {
                    name: value
                }
            }
            """
        )
        section = config_parser.Parser.parse(text.encode("utf-8"))
        self.assertEqual(config_parser.verify_section(section), ([], [], []))

    def test_bad_section(self):
        section = config_parser.Section("my#section")
        self.assertEqual(
            config_parser.verify_section(section), (["my#section"], [], [])
        )

    def test_bad_attr_name(self):
        section = config_parser.Section("mySection")
        section.add_attribute("bad#name", "value1")
        section.add_attribute("good_name", "value2")
        self.assertEqual(
            config_parser.verify_section(section),
            ([], ["mySection.bad#name"], []),
        )

    def test_bad_attr_value(self):
        section = config_parser.Section("mySection")
        section.add_attribute("bad_value", "va{l}ue1")
        section.add_attribute("good_value", "value2")
        self.assertEqual(
            config_parser.verify_section(section),
            ([], [], [("mySection.bad_value", "va{l}ue1")]),
        )

    def test_complex(self):
        text = outdent(
            """\
            name1: value1
            name#2: value2

            child1 {
                name1_1: value1.1
                name1#2: value1.2

                child1A {
                    name1A1: value
                }
                child1B# {
                    name#1B1: value
                    name1B2: value
                }
            }

            child2 {
                child2# {
                    na#me: value
                }
            }
            """
        )
        section = config_parser.Parser.parse(text.encode("utf-8"))
        # this would be rejected by the parser
        section.add_attribute("name1_3", "va{l}ue")
        self.assertEqual(
            config_parser.verify_section(section),
            (
                ["child1.child1B#", "child2.child2#"],
                [
                    "name#2",
                    "child1.name1#2",
                    "child1.child1B#.name#1B1",
                    "child2.child2#.na#me",
                ],
                [("name1_3", "va{l}ue")],
            ),
        )



from unittest import TestCase

from pcs.common import reports
from pcs.lib.commands import tag as cmd_tag
from pcs_test.tier0.lib.commands.tag.tag_common import (
    fixture_constraints_for_tags,
    fixture_resources_for_ids,
    fixture_tags_xml,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


TAG_DEFINITIONS = [
    ("tag1", ("i1", "i2")),
    ("tag2", ("j1", "j2")),
    ("tag3", ("k1", "k2")),
]


class TestTagRemove(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_tags_element_is_kept(self):
        self.config.runner.cib.load(
            tags=fixture_tags_xml(TAG_DEFINITIONS[0:1]),
        )
        self.config.env.push_cib(tags="<tags/>")
        cmd_tag.remove(self.env_assist.get_env(), ["tag1"])

    def test_remove_tags_others_are_kept(self):
        self.config.runner.cib.load(
            tags=fixture_tags_xml(TAG_DEFINITIONS),
        )
        self.config.env.push_cib(
            tags=fixture_tags_xml(TAG_DEFINITIONS[2:3]),
        )
        cmd_tag.remove(self.env_assist.get_env(), ["tag1", "tag2"])

    def test_remove_all_tags(self):
        self.config.runner.cib.load(tags=fixture_tags_xml(TAG_DEFINITIONS))
        self.config.env.push_cib(tags="<tags/>")
        cmd_tag.remove(self.env_assist.get_env(), ["tag1", "tag2", "tag3"])

    def test_nonexistent_tag_ids(self):
        self.config.runner.cib.load(tags=fixture_tags_xml(TAG_DEFINITIONS))
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.remove(
                self.env_assist.get_env(),
                ["nonexistent_tag1", "nonexistent_tag2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found(
                    _id, expected_types=["tag"], context_type="tags"
                )
                for _id in ["nonexistent_tag1", "nonexistent_tag2"]
            ]
        )

    def test_not_tag_ids(self):
        self.config.runner.cib.load(
            resources=fixture_resources_for_ids(),
            tags=fixture_tags_xml(TAG_DEFINITIONS),
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.remove(self.env_assist.get_env(), ["id1", "id2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.report_unexpected_element(
                    _id,
                    "primitive",
                    expected_types=["tag"],
                )
                for _id in ["id1", "id2"]
            ]
        )

    def test_mixed_ids(self):
        self.config.runner.cib.load(
            resources=fixture_resources_for_ids(),
            tags=fixture_tags_xml(TAG_DEFINITIONS),
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.remove(
                self.env_assist.get_env(),
                ["tag1", "nonexistent_tag1", "id1"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found(
                    "nonexistent_tag1",
                    expected_types=["tag"],
                    context_type="tags",
                ),
                fixture.report_unexpected_element(
                    "id1",
                    "primitive",
                    expected_types=["tag"],
                ),
            ]
        )

    def test_tag_referenced_in_constraint(self):
        self.config.runner.cib.load(
            constraints=fixture_constraints_for_tags(
                tag[0] for tag in TAG_DEFINITIONS
            ),
            tags=fixture_tags_xml(TAG_DEFINITIONS),
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.remove(
                self.env_assist.get_env(),
                ["tag1"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS,
                    tag_id="tag1",
                    constraint_id_list=["location-tag1"],
                ),
            ]
        )

    def test_multiple_tags_referenced_in_constraints(self):
        self.config.runner.cib.load(
            constraints=fixture_constraints_for_tags(
                tag[0] for tag in TAG_DEFINITIONS
            ),
            tags=fixture_tags_xml(TAG_DEFINITIONS),
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.remove(
                self.env_assist.get_env(),
                ["tag1", "tag2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS,
                    tag_id="tag1",
                    constraint_id_list=["location-tag1"],
                ),
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS,
                    tag_id="tag2",
                    constraint_id_list=["location-tag2"],
                ),
            ]
        )



import difflib
import doctest
import re
from lxml.doctestcompare import LXMLOutputChecker
from lxml.etree import LXML_VERSION

from pcs.common import reports
from pcs.lib.errors import LibraryError

# pylint: disable=invalid-name, no-self-use

# cover python2 vs. python3 differences
_re_object_type = type(re.compile(""))


def prepare_diff(first, second):
    """
    Return a string containing a diff of first and second
    """
    return "".join(
        difflib.Differ().compare(first.splitlines(1), second.splitlines(1))
    )


def ac(a, b):
    """
    Compare the actual output 'a' and an expected output 'b', print diff b a
    """
    if a != b:
        raise AssertionError(
            "strings not equal:\n{0}".format(prepare_diff(b, a))
        )


def start_tag_error_text():
    """lxml 3.7+ gives a longer 'start tag expected' error message,
    handle it here so multiple tests can just get the appropriate
    string from this function.
    """
    msg = "Start tag expected, '<' not found, line 1, column 1"
    if LXML_VERSION >= (3, 7, 0, 0):
        msg += " (<string>, line 1)"
    return msg


def console_report(*lines):
    # after lines append last new line
    return "\n".join(lines + ("",))


class AssertPcsMixin:
    """Run pcs command and assert its result"""

    def assert_pcs_success_all(self, command_list):
        for command in command_list:
            stdout, pcs_returncode = self.pcs_runner.run(command)
            if pcs_returncode != 0:
                raise AssertionError(
                    (
                        "Command '{0}' does not succeed.\n"
                        "return_code: {1}\n"
                        "stdout:\n{2}"
                    ).format(command, pcs_returncode, stdout)
                )

    def assert_pcs_success(
        self,
        command,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
        despace=False,
    ):
        full = stdout_full
        if (
            stdout_start is None
            and stdout_full is None
            and stdout_regexp is None
        ):
            full = ""
        self.assert_pcs_result(
            command,
            stdout_full=full,
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            returncode=0,
            despace=despace,
        )

    def assert_pcs_fail(
        self, command, stdout_full=None, stdout_start=None, stdout_regexp=None
    ):
        self.assert_pcs_result(
            command,
            stdout_full=stdout_full,
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            returncode=1,
        )

    def assert_pcs_fail_regardless_of_force(
        self, command, stdout_full=None, stdout_start=None, stdout_regexp=None
    ):
        self.assert_pcs_fail(command, stdout_full, stdout_start, stdout_regexp)
        self.assert_pcs_fail(
            command + ["--force"], stdout_full, stdout_start, stdout_regexp
        )

    def assert_pcs_result(
        self,
        command,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
        returncode=0,
        despace=False,
    ):
        msg = (
            "Please specify exactly one: stdout_start or stdout_full or"
            " stdout_regexp"
        )
        specified_stdout = [
            stdout
            for stdout in (stdout_full, stdout_start, stdout_regexp)
            if stdout is not None
        ]
        if not specified_stdout:
            raise Exception(msg + ", none specified")
        if len(specified_stdout) > 1:
            raise Exception(msg + ", both specified")

        stdout, pcs_returncode = self.pcs_runner.run(command)
        self.assertEqual(
            returncode,
            pcs_returncode,
            (
                'Expected return code "{0}", but was "{1}"'
                + "\ncommand: {2}\nstdout:\n{3}"
            ).format(returncode, pcs_returncode, command, stdout),
        )
        message_template = (
            "{reason}\ncommand: {cmd}\ndiff is (expected is 2nd):\n{diff}"
            + "\nFull stdout:\n{stdout}"
        )
        if stdout_start:
            expected_start = self.__prepare_output(stdout_start)
            if not stdout.startswith(expected_start):
                self.fail(
                    message_template.format(
                        reason="Stdout does not start as expected",
                        cmd=command,
                        diff=prepare_diff(
                            stdout[: len(expected_start)], expected_start
                        ),
                        stdout=stdout,
                    )
                )
        elif stdout_regexp:
            if not isinstance(stdout_regexp, _re_object_type):
                stdout_regexp = re.compile(stdout_regexp)
            if not stdout_regexp.search(stdout):
                self.fail(
                    (
                        "Stdout does not match the expected regexp\n"
                        "command: {cmd}\nregexp:\n{regexp} (flags: {flags})\n"
                        "\nFull stdout:\n{stdout}"
                    ).format(
                        cmd=command,
                        regexp=stdout_regexp.pattern,
                        flags=", ".join(
                            self.__prepare_regexp_flags(stdout_regexp.flags)
                        ),
                        stdout=stdout,
                    )
                )
        else:
            expected_full = self.__prepare_output(stdout_full)
            if (despace and _despace(stdout) != _despace(expected_full)) or (
                not despace and stdout != expected_full
            ):
                self.assertEqual(
                    stdout,
                    expected_full,
                    message_template.format(
                        reason="Stdout is not as expected",
                        cmd=command,
                        diff=prepare_diff(stdout, expected_full),
                        stdout=stdout,
                    ),
                )

    def __prepare_output(self, output):
        if isinstance(output, list):
            return console_report(*output)
        return output

    def __prepare_regexp_flags(self, flags):
        # python2 has different flags than python3
        possible_flags = [
            "ASCII",
            "DEBUG",
            "IGNORECASE",
            "LOCALE",
            "MULTILINE",
            "DOTALL",
            "UNICODE",
            "VERBOSE",
        ]
        used_flags = [
            f
            for f in possible_flags
            if hasattr(re, f) and (flags & getattr(re, f))
        ]
        return sorted(used_flags)


class ExtendedAssertionsMixin:
    def assert_raises(
        self, expected_exception, callable_obj, property_dict=None
    ):
        if property_dict is None:
            property_dict = {}
        try:
            callable_obj()
            raise AssertionError(
                "No exception raised. Expected exception: {exception}".format(
                    exception=expected_exception.__class__.__name__
                )
            )
        except expected_exception as e:
            for prop, value in property_dict.items():
                try:
                    self.assertEqual(value, getattr(e, prop))
                except AttributeError as e:
                    raise AssertionError(
                        "Property {property} doesn't exist in exception"
                        " {exception}".format(
                            property=prop, exception=e.__class__.__name__
                        )
                    ) from e


def assert_xml_equal(expected_xml, got_xml, context_explanation=""):
    checker = LXMLOutputChecker()
    if not checker.check_output(expected_xml, got_xml, 0):
        raise AssertionError(
            "{context_explanation}{xml_diff}".format(
                context_explanation=(
                    ""
                    if not context_explanation
                    else "\n{0}\n".format(context_explanation)
                ),
                xml_diff=checker.output_difference(
                    doctest.Example("", expected_xml), got_xml, 0
                ),
            )
        )


SEVERITY_SHORTCUTS = {
    reports.ReportItemSeverity.INFO: "I",
    reports.ReportItemSeverity.WARNING: "W",
    reports.ReportItemSeverity.ERROR: "E",
    reports.ReportItemSeverity.DEBUG: "D",
}


def _format_report_item_info(info):
    return ", ".join(
        ["{0}:{1}".format(key, repr(value)) for key, value in info.items()]
    )


def _expected_report_item_format(report_item_expectation):
    return "{0} {1} {{{2}}} ! {3} {4}".format(
        SEVERITY_SHORTCUTS.get(
            report_item_expectation[0], report_item_expectation[0]
        ),
        report_item_expectation[1],
        _format_report_item_info(report_item_expectation[2]),
        report_item_expectation[3]
        if len(report_item_expectation) > 3
        else None,
        report_item_expectation[4]
        if len(report_item_expectation) > 4
        else None,
    )


def _format_report_item(report_item):
    return _expected_report_item_format(
        (
            report_item.severity.level,
            report_item.message.code,
            report_item.message.to_dto().payload,
            report_item.severity.force_code,
        )
    )


def assert_report_item_equal(real_report_item, report_item_info):
    if not __report_item_equal(real_report_item, report_item_info):
        raise AssertionError(
            "ReportItem not equal\nexpected: {0}\nactual:   {1}".format(
                repr(
                    (
                        report_item_info[0],
                        report_item_info[1],
                        report_item_info[2],
                        None
                        if len(report_item_info) < 4
                        else report_item_info[3],
                        None
                        if len(report_item_info) < 5
                        else report_item_info[4],
                    )
                ),
                _format_report_item(real_report_item),
            )
        )


def _unexpected_report_given(
    remaining_expected_report_info_list,
    expected_report_info_list,
    real_report_item,
    real_report_item_list,
):
    return AssertionError(
        (
            "\n  Unexpected real report given:"
            "\n  =============================\n    {0}\n"
            "\n  remaining expected reports ({1}) are:"
            "\n  ------------------------------------\n    {2}\n"
            "\n  all expected reports ({3}) are:"
            "\n  ------------------------------\n    {4}\n"
            "\n  all real reports ({5}):"
            "\n  ---------------------\n    {6}"
        ).format(
            _format_report_item(real_report_item),
            len(remaining_expected_report_info_list),
            "\n    ".join(
                map(
                    _expected_report_item_format,
                    remaining_expected_report_info_list,
                )
            )
            if remaining_expected_report_info_list
            else "No other report is expected!",
            len(expected_report_info_list),
            "\n    ".join(
                map(
                    _expected_report_item_format,
                    expected_report_info_list,
                )
            )
            if expected_report_info_list
            else "No report is expected!",
            len(real_report_item_list),
            "\n    ".join(map(_format_report_item, real_report_item_list)),
        )
    )


def assert_report_item_list_equal(
    real_report_item_list, expected_report_info_list, hint=""
):
    remaining_expected_report_info_list = expected_report_info_list[:]
    duplicate_report_item_is_missing = False
    for real_report_item in real_report_item_list:
        found_report_info = __find_report_info(
            expected_report_info_list, real_report_item
        )
        if found_report_info is None:
            raise _unexpected_report_given(
                remaining_expected_report_info_list,
                expected_report_info_list,
                real_report_item,
                real_report_item_list,
            )
        if found_report_info in remaining_expected_report_info_list:
            remaining_expected_report_info_list.remove(found_report_info)
        else:
            duplicate_report_item_is_missing = True
    if remaining_expected_report_info_list or duplicate_report_item_is_missing:

        def format_items(item_type, item_list):
            caption = "{0} ReportItems({1})".format(item_type, len(item_list))
            return "{0}\n{1}\n{2}".format(
                caption, "-" * len(caption), "\n".join(map(repr, item_list))
            )

        raise AssertionError(
            "\nExpected LibraryError is missing\n{0}\n\n{1}\n\n{2}".format(
                "{0}\n".format(hint) if hint else "",
                format_items("expected", expected_report_info_list),
                format_items("real", real_report_item_list),
            )
        )


def assert_raise_library_error(callableObj, *report_info_list):
    try:
        callableObj()
        raise AssertionError("LibraryError not raised")
    except LibraryError as e:
        assert_report_item_list_equal(e.args, list(report_info_list))


def __find_report_info(expected_report_info_list, real_report_item):
    for report_info in expected_report_info_list:
        if __report_item_equal(real_report_item, report_info):
            return report_info
    return None


def __report_item_equal(real_report_item, report_item_info):
    report_dto: reports.ReportItemDto = real_report_item.to_dto()
    return (
        report_dto.severity.level == report_item_info[0]
        and report_dto.message.code == report_item_info[1]
        and report_dto.message.payload == report_item_info[2]
        and (
            report_dto.severity.force_code
            == (None if len(report_item_info) < 4 else report_item_info[3])
        )
        and report_dto.context
        == (report_item_info[4] if len(report_item_info) >= 5 else None)
    )


def assert_pcs_status(status1, status2):
    if _despace(status1) != _despace(status2):
        raise AssertionError(
            "strings not equal:\n{0}".format(prepare_diff(status1, status2))
        )


def _despace(string):
    # ignore whitespace changes between various pacemaker versions
    return re.sub(r"[ \t]+", " ", string)



from typing import Any, Dict, Iterable, List, Optional, Type

from pcs.common.interface.dto import to_dict
from pcs.lib import resource_agent
from pcs.lib.env import LibraryEnvironment
from pcs.lib.external import CommandRunner


def list_standards(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents standards (ocf, lsb, ... ) on the local host
    """
    return resource_agent.list_resource_agents_standards(lib_env.cmd_runner())


def list_ocf_providers(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents ocf providers on the local host
    """
    return resource_agent.list_resource_agents_ocf_providers(
        lib_env.cmd_runner()
    )


def list_agents_for_standard_and_provider(
    lib_env: LibraryEnvironment, standard_provider: Optional[str] = None
) -> List[str]:
    """
    List resource agents for specified standard on the local host

    standard_provider -- standard[:provider], e.g. None, ocf, ocf:pacemaker
    """
    if standard_provider:
        standards = [standard_provider]
    else:
        standards = resource_agent.list_resource_agents_standards(
            lib_env.cmd_runner()
        )
    agents = []
    for std in standards:
        agents += resource_agent.list_resource_agents(lib_env.cmd_runner(), std)
    return sorted(agents, key=str.lower)


# TODO return a list of DTOs
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all resource agents on the local host, optionally filtered and
        described

    describe -- load and return agents' description as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()

    # list agents for all standards and providers
    agent_names = []
    for std in resource_agent.list_resource_agents_standards_and_providers(
        runner
    ):
        agent_names += [
            "{0}:{1}".format(std, agent)
            for agent in resource_agent.list_resource_agents(runner, std)
        ]
    return _complete_agent_list(
        runner,
        sorted(agent_names, key=str.lower),
        describe,
        search,
        resource_agent.ResourceAgent,
    )


def _complete_agent_list(
    runner: CommandRunner,
    agent_names: Iterable[str],
    describe: bool,
    search: Optional[str],
    metadata_class: Type[resource_agent.CrmAgent],
) -> List[Dict[str, Any]]:
    # filter agents by name if requested
    if search:
        search_lower = search.lower()
        agent_names = [
            name for name in agent_names if search_lower in name.lower()
        ]

    # complete the output and load descriptions if requested
    agent_list = []
    for name in agent_names:
        try:
            agent_metadata = metadata_class(runner, name)
            metadata_dto = (
                agent_metadata.get_full_info()
                if describe
                else agent_metadata.get_name_info()
            )
            agent_list.append(to_dict(metadata_dto))
        except resource_agent.ResourceAgentError:
            # we don't return it in the list:
            #
            # UnableToGetAgentMetadata - if we cannot get valid metadata, it's
            # not a resource agent
            #
            # InvalidResourceAgentName - invalid name cannot be used with a new
            # resource. The list of names is gained from "crm_resource" whilst
            # pcs is doing the validation. So there can be a name that pcs does
            # not recognize as valid.
            #
            # Providing a warning is not the way (currently). Other components
            # read this list and do not expect warnings there. Using the stderr
            # (to separate warnings) is currently difficult.
            pass
    return agent_list


# TODO return aDTO
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent
    """
    agent = resource_agent.find_valid_resource_agent_by_name(
        lib_env.report_processor,
        lib_env.cmd_runner(),
        agent_name,
        absent_agent_supported=False,
    )
    return to_dict(agent.get_full_info())



import os.path

from pcs.daemon import session
from pcs.daemon.app import session as app_session
from pcs.daemon.app.common import BaseHandler
from pcs.daemon.app.ui_common import AjaxMixin, StaticFile


class SPAHandler(BaseHandler):
    __index = None
    __fallback = None

    def initialize(self, index, fallback):
        # pylint: disable=arguments-differ
        self.__index = index
        self.__fallback = fallback
        # add security related headers to all responses
        self.enhance_headers()

    def get(self, *args, **kwargs):
        del args, kwargs
        self.render(
            self.__index
            if os.path.isfile(str(self.__index))
            # spa is probably not installed
            else self.__fallback
        )


class Login(SPAHandler, app_session.Mixin, AjaxMixin):
    def initialize(self, session_storage, index, fallback):
        # pylint: disable=arguments-differ
        app_session.Mixin.initialize(self, session_storage)
        SPAHandler.initialize(self, index, fallback)

    async def post(self, *args, **kwargs):
        del args, kwargs
        # This is the way of old (ruby) pcsd. Post login generates a session
        # cookie. No matter if authentication succeeded or failed.

        await self.session_auth_user(
            self.get_body_argument("username"),
            self.get_body_argument("password"),
            sign_rejection=False,
        )

        if not self.session.is_authenticated:
            raise self.unauthorized()

        self.write(self.session.ajax_id)


class Logout(app_session.Mixin, AjaxMixin, BaseHandler):
    """
    Logout handles url for logout. It is used for both ajax and non-ajax
    requests.
    """

    async def get(self, *args, **kwargs):
        del args, kwargs
        await self.init_session()
        self.session_logout()
        self.sid_to_cookies()
        self.enhance_headers()
        self.write("OK")


class StaticFileMayBe(StaticFile):
    # pylint: disable=abstract-method
    def get(self, *args, **kwargs):
        # pylint: disable=signature-differs
        # pylint: disable=invalid-overridden-method
        if not os.path.isdir(str(self.root)):
            # spa is probably not installed
            self.set_status(404, "Not Found")
            return None
        return super().get(*args, **kwargs)


def get_routes(
    url_prefix,
    app_dir,
    fallback_page_path,
    session_storage: session.Storage,
):
    sessions = dict(session_storage=session_storage)
    static_path = lambda dir="": dict(path=os.path.join(app_dir, dir))
    pages = dict(
        index=os.path.join(app_dir, "index.html"),
        fallback=fallback_page_path,
    )

    # One possibility is to check if SPA index exists and if does not exists
    # return only route to fallback page. However, it would require to restart
    # pcsd after putting spa. For this reason, routes stay the same, no matter
    # if SPA index exists.

    return [
        (f"{url_prefix}static/(.*)", StaticFileMayBe, static_path("static")),
        (f"{url_prefix}manifest.json", StaticFileMayBe, static_path()),
        (f"{url_prefix}login", Login, {**sessions, **pages}),
        (f"{url_prefix}logout", Logout, sessions),
        (f"{url_prefix}.*", SPAHandler, pages),
    ]



import inspect
import logging
import os
import os.path
import sys
from functools import partial

from unittest import mock

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error, prepare_diff
from pcs_test.tools.case_analysis import test_failed
from pcs_test.tools.command_env import spy
from pcs_test.tools.command_env.calls import Queue as CallQueue
from pcs_test.tools.command_env.config import Config
from pcs_test.tools.command_env.mock_fs import (
    get_fs_mock,
    is_fs_call_in,
)
from pcs_test.tools.command_env.mock_get_local_corosync_conf import (
    get_get_local_corosync_conf,
)
from pcs_test.tools.command_env.mock_node_communicator import NodeCommunicator
from pcs_test.tools.command_env.mock_push_cib import (
    get_push_cib,
    is_push_cib_call_in,
)
from pcs_test.tools.command_env.mock_push_corosync_conf import (
    get_push_corosync_conf,
    is_push_corosync_conf_call_in,
)
from pcs_test.tools.command_env.mock_raw_file import get_raw_file_mock
from pcs_test.tools.command_env.mock_runner import Runner
from pcs_test.tools.command_env.mock_service_manager import ServiceManagerMock
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common.file import RawFile
from pcs.common.node_communicator import NodeCommunicatorFactory
from pcs.lib.env import LibraryEnvironment

patch_lib_env = partial(mock.patch.object, LibraryEnvironment)


def patch_env(call_queue, config, init_env, is_systemd=True):
    # pylint: disable=too-many-locals
    # It is mandatory to patch some env objects/methods. It is ok when command
    # does not use this objects/methods and specify no call for it. But it would
    # be a problem when the test succeded because the live call respond
    # correctly by accident. Such test would fails on different machine (with
    # another live environment)

    orig_cmd_runner = init_env.cmd_runner
    get_node_communicator = init_env.get_node_communicator
    mock_communicator_factory = mock.Mock(spec_set=NodeCommunicatorFactory)
    mock_communicator_factory.get_communicator = (
        # TODO: use request_timeout
        lambda request_timeout=None: NodeCommunicator(call_queue)
        if not config.spy
        else spy.NodeCommunicator(get_node_communicator())
    )

    def get_cmd_runner(self, env=None):
        del self
        if config.spy:
            return spy.Runner(orig_cmd_runner())
        env_vars = {}
        if config.env.cib_tempfile:
            env_vars["CIB_file"] = config.env.cib_tempfile
        if env:
            env_vars.update(env)
        return Runner(call_queue, env_vars=env_vars)

    patcher_list = [
        patch_lib_env("cmd_runner", get_cmd_runner),
        mock.patch(
            "pcs.lib.env.get_local_corosync_conf",
            get_get_local_corosync_conf(call_queue)
            if not config.spy
            else spy.get_local_corosync_conf,
        ),
        patch_lib_env("communicator_factory", mock_communicator_factory),
        # Use our custom ServiceManager in tests
        # TODO: add support for Spy
        patch_lib_env(
            "_get_service_manager", lambda _: ServiceManagerMock(call_queue)
        ),
    ]
    if is_systemd:
        # In most test cases we don't care about underlaying init system. But
        # some tests actually test different behavior based on init
        # system. This will cause pcs.lib.services.is_systemd(<instance of
        # ServiceManagerMock>) to return True because we replace SystemdDriver
        # in isinstance() call inside is_systemd function with
        # ServiceManagerMock. Otherwise is_systemd will return False.
        patcher_list.append(
            mock.patch(
                "pcs.lib.services.services.drivers.SystemdDriver",
                ServiceManagerMock,
            )
        )

    if is_fs_call_in(call_queue):
        fs_mock = get_fs_mock(call_queue)
        builtin = (
            ("__builtin__" if sys.version_info[0] == 2 else "builtins") + ".{0}"
        ).format

        patcher_list.extend(
            [
                mock.patch(builtin("open"), fs_mock("open", open)),
                mock.patch(
                    "os.path.exists", fs_mock("os.path.exists", os.path.exists)
                ),
                mock.patch(
                    "os.path.isdir", fs_mock("os.path.isdir", os.path.isdir)
                ),
                mock.patch(
                    "os.path.isfile", fs_mock("os.path.isfile", os.path.isfile)
                ),
                mock.patch("os.listdir", fs_mock("os.listdir", os.listdir)),
                mock.patch("os.chmod", fs_mock("os.chmod", os.chmod)),
                mock.patch("os.chown", fs_mock("os.chown", os.chown)),
            ]
        )

    raw_file_mock = get_raw_file_mock(call_queue)
    for method_name, dummy_method in inspect.getmembers(
        RawFile, inspect.isfunction
    ):
        # patch all public methods
        # inspect.isfunction must be used instead of ismethod because we are
        # working with a class and not an instance - no method is bound yet so
        # it would return an empty list
        # "protected" methods start with _
        # "private" methods start with _<class_name>__
        if method_name.startswith("_"):
            continue
        patcher_list.append(
            mock.patch.object(
                RawFile, method_name, getattr(raw_file_mock, method_name)
            )
        )

    # It is not always desirable to patch these methods. Some tests may patch
    # only the internals (runner etc.). So these methods are only patched when
    # it is explicitly configured.
    if is_push_cib_call_in(call_queue):
        patcher_list.append(patch_lib_env("push_cib", get_push_cib(call_queue)))
    if is_push_corosync_conf_call_in(call_queue):
        patcher_list.append(
            patch_lib_env(
                "push_corosync_conf", get_push_corosync_conf(call_queue)
            )
        )

    for patcher in patcher_list:
        patcher.start()

    def unpatch():
        for patcher in patcher_list:
            patcher.stop()

    return unpatch


class EnvAssistant:
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        config=None,
        test_case=None,
        exception_reports_in_processor_by_default=True,
    ):
        """
        TestCase test_case -- cleanup callback is registered to test_case if is
            provided
        """
        self.__call_queue = None
        self.__config = config if config else Config()
        self.__reports_asserted = False
        self.__extra_reports = []
        self.exception_reports_in_processor_by_default = (
            exception_reports_in_processor_by_default
        )

        self.__unpatch = None
        self.__original_mocked_corosync_conf = None

        if test_case:
            test_case.addCleanup(lambda: self.cleanup(test_case))

    @property
    def config(self):
        return self.__config

    def cleanup(self, current_test):
        if self.__unpatch:
            self.__unpatch()

        if test_failed(current_test):
            # We have already got the message that main test failed. There is
            # a high probability that something remains in reports or in the
            # queue etc. But it is only consequence of the main test fail. And
            # we do not want to make the report confusing.
            return

        if not self.__reports_asserted:
            self.__assert_environment_created()
            if not self.__config.spy:
                self._env.report_processor.assert_reports(
                    self.__extra_reports,
                    hint="EnvAssistant.cleanup - is param"
                    " 'expected_in_processor' in the method"
                    " 'assert_raise_library_error' set correctly?",
                )

        if not self.__config.spy:
            if self.__call_queue and self.__call_queue.remaining:
                raise AssertionError(
                    "There are remaining expected calls: \n    '{0}'".format(
                        "'\n    '".join(
                            [repr(call) for call in self.__call_queue.remaining]
                        )
                    )
                )
            # If pushing corosync.conf has not been patched in the
            # LibraryEnvironment and the LibraryEnvironment was constructed
            # with a mocked corosync.conf, check if it was changed without the
            # change being specified in a test.
            # If no env.push_corosync_conf call has been specified, no mocking
            # occurs, any changes to corosync.conf are done just in memory and
            # nothing gets reported. So an explicit check is necessary.
            corosync_conf_orig = self.__original_mocked_corosync_conf
            # pylint: disable=protected-access
            corosync_conf_env = self._env._corosync_conf_data
            if corosync_conf_orig and corosync_conf_orig != corosync_conf_env:
                raise AssertionError(
                    (
                        "An unexpected change to corosync.conf in "
                        "LibraryEnvironment has been detected:\n{0}"
                    ).format(
                        prepare_diff(corosync_conf_orig, corosync_conf_env)
                    )
                )

    def get_env(self, is_systemd=True):
        self.__call_queue = CallQueue(self.__config.calls)
        # pylint: disable=attribute-defined-outside-init
        self._env = LibraryEnvironment(
            mock.MagicMock(logging.Logger),
            MockLibraryReportProcessor(),
            cib_data=self.__config.env.cib_data,
            corosync_conf_data=self.__config.env.corosync_conf_data,
            known_hosts_getter=(
                (lambda: self.__config.spy.known_hosts)
                if self.__config.spy
                else self.__config.env.known_hosts_getter
            ),
            booth_files_data=self.__config.env.booth,
        )
        self.__unpatch = patch_env(
            self.__call_queue,
            self.__config,
            self._env,
            is_systemd=is_systemd,
        )
        # If pushing corosync.conf has not been patched in the
        # LibraryEnvironment, store any corosync.conf passed to the
        # LibraryEnvironment for check for changes in cleanup.
        if not is_push_corosync_conf_call_in(self.__call_queue):
            self.__original_mocked_corosync_conf = (
                self.__config.env.corosync_conf_data
            )
        return self._env

    def assert_reports(self, expected_reports):
        self.__reports_asserted = True
        self.__assert_environment_created()
        self._env.report_processor.assert_reports(
            (
                expected_reports.reports
                if isinstance(expected_reports, fixture.ReportStore)
                else expected_reports
            )
            + self.__extra_reports
        )

    def assert_raise_library_error(
        self, command, reports=None, expected_in_processor=None
    ):
        if reports is None:
            reports = []

        if not isinstance(reports, list):
            raise self.__list_of_reports_expected(reports)

        if expected_in_processor is None:
            expected_in_processor = (
                self.exception_reports_in_processor_by_default
            )

        assert_raise_library_error(command, *reports)
        if expected_in_processor:
            self.__extra_reports = reports

    def __assert_environment_created(self):
        if not hasattr(self, "_env"):
            raise AssertionError(
                "LibraryEnvironment was not created in EnvAssitant."
                " Have you called method get_env?"
            )

    def __list_of_reports_expected(self, reports):
        return AssertionError(
            (
                "{0}.{1} expects 'list' as reports parameter, '{2}' was given"
            ).format(
                self.__class__.__name__, "assert_raise", type(reports).__name__
            )
        )



import os.path
from typing import Iterable

from pcs import settings
from pcs.common import reports
from pcs.common.str_tools import join_multilines
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError


def unfence_node(
    env: LibraryEnvironment,
    node: str,
    original_devices: Iterable[str],
    updated_devices: Iterable[str],
) -> None:
    """
    Unfence scsi devices on a node by calling fence_scsi agent script. Only
    newly added devices will be unfenced (set(updated_devices) -
    set(original_devices)). Before unfencing, original devices are be checked
    if any of them are not fenced. If there is a fenced device, unfencing will
    be skipped.

    env -- provides communication with externals
    node -- node name on wich is unfencing performed
    original_devices -- list of devices defined before update
    updated_devices -- list of devices defined after update
    """
    devices_to_unfence = set(updated_devices) - set(original_devices)
    if not devices_to_unfence:
        return
    fence_scsi_bin = os.path.join(settings.fence_agent_binaries, "fence_scsi")
    fenced_devices = []
    for device in original_devices:
        stdout, stderr, return_code = env.cmd_runner().run(
            [
                fence_scsi_bin,
                "--action=status",
                f"--devices={device}",
                f"--plug={node}",
            ]
        )
        if return_code == 2:
            fenced_devices.append(device)
        elif return_code != 0:
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.StonithUnfencingDeviceStatusFailed(
                        device, join_multilines([stderr, stdout])
                    )
                )
            )
    if fenced_devices:
        # At least one of existing devices is off, which means the node has
        # been fenced and new devices should not be unfenced.
        env.report_processor.report(
            reports.ReportItem.info(
                reports.messages.StonithUnfencingSkippedDevicesFenced(
                    fenced_devices
                )
            )
        )
        return
    stdout, stderr, return_code = env.cmd_runner().run(
        [
            fence_scsi_bin,
            "--action=on",
            "--devices",
            ",".join(sorted(devices_to_unfence)),
            f"--plug={node}",
        ],
    )
    if return_code != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.StonithUnfencingFailed(
                    join_multilines([stderr, stdout])
                )
            )
        )



from functools import partial

from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.constraint import constraint
from pcs.lib.cib.tools import check_new_id_applicable
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import is_score, SCORE_INFINITY

TAG_NAME = "rsc_colocation"
DESCRIPTION = "constraint id"


def prepare_options_with_set(cib, options, resource_set_list):
    options = constraint.prepare_options(
        ("score",),
        options,
        partial(constraint.create_id, cib, "colocation", resource_set_list),
        partial(check_new_id_applicable, cib, DESCRIPTION),
    )

    if "score" in options:
        if not is_score(options["score"]):
            raise LibraryError(
                ReportItem.error(
                    reports.messages.InvalidScore(options["score"])
                )
            )
    else:
        options["score"] = SCORE_INFINITY

    return options



from pcs_test.tools.command_env.mock_raw_file import (
    RawFileExistsCall,
    RawFileReadCall,
    RawFileRemoveCall,
    RawFileWriteCall,
)


class RawFileConfig:
    def __init__(self, call_collection):
        self.__calls = call_collection

    def exists(
        self,
        file_type_code,
        path,
        exists=True,
        name="raw_file.exists",
        before=None,
        instead=None,
    ):
        """
        Create a call for checking a file existence

        string file_type_code -- item from pcs.common.file_type_codes
        string path -- expected file path
        bool exists -- the result of the existence check
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        call = RawFileExistsCall(file_type_code, path, exists=exists)
        self.__calls.place(name, call, before, instead)

    def read(
        self,
        file_type_code,
        path,
        content=None,
        exception_msg=None,
        name="raw_file.read",
        before=None,
        instead=None,
    ):
        """
        Create a call for reading a file content

        string file_type_code -- item from pcs.common.file_type_codes
        string path -- expected file path
        bytes content -- the result of succesful reading
        string exception_msg -- resulting error in case of unsuccessful reading
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        call = RawFileReadCall(
            file_type_code,
            path,
            content=content,
            exception_msg=exception_msg,
        )
        self.__calls.place(name, call, before, instead)

    def write(
        self,
        file_type_code,
        path,
        file_data,
        can_overwrite=False,
        already_exists=False,
        exception_msg=None,
        exception_action=None,
        name="raw_file.write",
        before=None,
        instead=None,
    ):
        """
        Create a call for writing to a file

        string file_type_code -- item from pcs.common.file_type_codes
        string path -- expected file path
        bytes file_data -- expected data to be written
        bool already_exist -- result is that the file already exists
        string exception_msg -- resulting error in case of unsuccessful write
        string exception_action -- item of pcs.common.file.RawFileError
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        # pylint: disable=too-many-arguments
        call = RawFileWriteCall(
            file_type_code,
            path,
            file_data,
            can_overwrite=can_overwrite,
            already_exists=already_exists,
            exception_msg=exception_msg,
            exception_action=exception_action,
        )
        self.__calls.place(name, call, before, instead)

    def remove(
        self,
        file_type_code,
        path,
        fail_if_file_not_found=True,
        exception_msg=None,
        file_not_found_exception=False,
        name="raw_file.remove",
        before=None,
        instead=None,
    ):
        """
        Create a call for removing a file

        string file_type_code -- item from pcs.common.file_type_codes
        string path -- expected file path
        bool fail_if_file_not_found -- should it fail when file not found?
        string exception_msg -- resulting error in case of unsuccessful removal
        bool file_not_found_exception -- raise an exception file was not found
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        # pylint: disable=too-many-arguments
        call = RawFileRemoveCall(
            file_type_code,
            path,
            fail_if_file_not_found=fail_if_file_not_found,
            exception_msg=exception_msg,
            file_not_found_exception=file_not_found_exception,
        )
        self.__calls.place(name, call, before, instead)



from unittest import TestCase

from lxml import etree

from pcs.common.interface import dto
from pcs.common.pacemaker.resource.relations import (
    RelationEntityDto,
    ResourceRelationDto,
    ResourceRelationType,
)
from pcs.lib.cib.resource import relations as lib


def fixture_cib(resources, constraints):
    return etree.fromstring(
        f"""
        <cib>
          <configuration>
            <resources>
            {resources}
            </resources>
            <constraints>
            {constraints}
            </constraints>
          </configuration>
        </cib>
    """
    )


def fixture_dummy_metadata(_id):
    return {
        "id": _id,
        "class": "c",
        "provider": "pcmk",
        "type": "Dummy",
    }


class ResourceRelationNode(TestCase):
    @staticmethod
    def entity_fixture(index):
        return dto.from_dict(
            RelationEntityDto,
            dict(
                id=f"ent_id{index}",
                type=str(ResourceRelationType.INNER_RESOURCES.value),
                members=[f"{index}m1", f"{index}m2", f"{index}m0"],
                metadata=dict(
                    id=f"ent_id{index}",
                    k0="val0",
                    k1="val1",
                ),
            ),
        )

    def assert_dto_equal(self, expected, actual):
        self.assertEqual(dto.to_dict(expected), dto.to_dict(actual))

    def test_no_members(self):
        ent = self.entity_fixture("0")
        obj = lib.ResourceRelationNode(ent)
        self.assert_dto_equal(ResourceRelationDto(ent, [], False), obj.to_dto())

    def test_with_members(self):
        ent0 = self.entity_fixture("0")
        ent1 = self.entity_fixture("1")
        ent2 = self.entity_fixture("2")
        ent3 = self.entity_fixture("3")
        obj = lib.ResourceRelationNode(ent0)
        obj.add_member(lib.ResourceRelationNode(ent1))
        member = lib.ResourceRelationNode(ent2)
        member.add_member(lib.ResourceRelationNode(ent3))
        obj.add_member(member)
        self.assert_dto_equal(
            ResourceRelationDto(
                ent0,
                [
                    ResourceRelationDto(ent1, [], False),
                    ResourceRelationDto(
                        ent2,
                        [
                            ResourceRelationDto(ent3, [], False),
                        ],
                        False,
                    ),
                ],
                False,
            ),
            obj.to_dto(),
        )

    def test_stop(self):
        ent = self.entity_fixture("0")
        obj = lib.ResourceRelationNode(ent)
        obj.stop()
        self.assert_dto_equal(
            ResourceRelationDto(ent, [], True),
            obj.to_dto(),
        )

    def test_add_member(self):
        ent0 = self.entity_fixture("0")
        ent1 = self.entity_fixture("1")
        obj = lib.ResourceRelationNode(ent0)
        obj.add_member(lib.ResourceRelationNode(ent1))
        self.assert_dto_equal(
            ResourceRelationDto(
                ent0,
                [ResourceRelationDto(ent1, [], False)],
                False,
            ),
            obj.to_dto(),
        )

    def test_add_member_itself(self):
        ent = self.entity_fixture("0")
        obj = lib.ResourceRelationNode(ent)
        obj.add_member(obj)
        self.assert_dto_equal(
            ResourceRelationDto(ent, [], False),
            obj.to_dto(),
        )

    def test_add_member_already_have_parent(self):
        obj0 = lib.ResourceRelationNode(self.entity_fixture("0"))
        obj1 = lib.ResourceRelationNode(self.entity_fixture("1"))
        obj2 = lib.ResourceRelationNode(self.entity_fixture("2"))
        obj0.add_member(obj1)
        with self.assertRaises(AssertionError):
            obj2.add_member(obj1)

    def test_add_member_already_in_branch(self):
        ent0 = self.entity_fixture("0")
        ent1 = self.entity_fixture("1")
        obj0 = lib.ResourceRelationNode(ent0)
        obj1 = lib.ResourceRelationNode(ent1)
        obj0.add_member(obj1)
        obj1.add_member(obj0)
        self.assert_dto_equal(
            ResourceRelationDto(
                ent0, [ResourceRelationDto(ent1, [], False)], False
            ),
            obj0.to_dto(),
        )

    def test_add_member_already_in_different_branch(self):
        ent0 = self.entity_fixture("0")
        ent1 = self.entity_fixture("1")
        obj0 = lib.ResourceRelationNode(ent0)
        obj0.add_member(lib.ResourceRelationNode(ent1))
        obj0.add_member(lib.ResourceRelationNode(ent1))
        self.assert_dto_equal(
            ResourceRelationDto(
                ent0,
                [
                    ResourceRelationDto(ent1, [], False),
                    ResourceRelationDto(ent1, [], False),
                ],
                False,
            ),
            obj0.to_dto(),
        )


class ResourceRelationsFetcher(TestCase):
    def test_ordering_constraint(self):
        obj = lib.ResourceRelationsFetcher(
            fixture_cib(
                """
            <primitive id="d1" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d2" class="c" provider="pcmk" type="Dummy"/>
            """,
                """
            <rsc_order first="d1" first-action="start"
                id="order-d1-d2-mandatory" then="d2" then-action="start"
                kind="Mandatory"/>
            """,
            )
        )
        expected = (
            {
                "d1": RelationEntityDto(
                    "d1",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["order-d1-d2-mandatory"],
                    fixture_dummy_metadata("d1"),
                ),
                "d2": RelationEntityDto(
                    "d2",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["order-d1-d2-mandatory"],
                    fixture_dummy_metadata("d2"),
                ),
            },
            {
                "order-d1-d2-mandatory": RelationEntityDto(
                    "order-d1-d2-mandatory",
                    ResourceRelationType.ORDER,
                    members=["d1", "d2"],
                    metadata={
                        "id": "order-d1-d2-mandatory",
                        "first": "d1",
                        "first-action": "start",
                        "then": "d2",
                        "then-action": "start",
                        "kind": "Mandatory",
                    },
                ),
            },
        )
        for res in ("d1", "d2"):
            with self.subTest(resource=res):
                self.assertEqual(expected, obj.get_relations(res))

    def test_ordering_set_constraint(self):
        obj = lib.ResourceRelationsFetcher(
            fixture_cib(
                """
            <primitive id="d1" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d2" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d3" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d4" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d5" class="c" provider="pcmk" type="Dummy"/>
            <primitive id="d6" class="c" provider="pcmk" type="Dummy"/>
            """,
                """
            <rsc_order kind="Serialize" symmetrical="true"
                id="pcs_rsc_order_set_1">
              <resource_set sequential="true" require-all="true" action="start"
                  id="pcs_rsc_set_1">
                <resource_ref id="d1"/>
                <resource_ref id="d3"/>
                <resource_ref id="d2"/>
              </resource_set>
              <resource_set action="stop" sequential="false"
                  require-all="false" id="pcs_rsc_set_2">
                <resource_ref id="d6"/>
                <resource_ref id="d5"/>
                <resource_ref id="d4"/>
              </resource_set>
            </rsc_order>
            """,
            )
        )
        rsc_entity = lambda _id: RelationEntityDto(
            _id,
            ResourceRelationType.RSC_PRIMITIVE,
            ["pcs_rsc_order_set_1"],
            fixture_dummy_metadata(_id),
        )
        res_list = ("d1", "d2", "d3", "d4", "d5", "d6")
        expected = (
            {_id: rsc_entity(_id) for _id in res_list},
            {
                "pcs_rsc_order_set_1": RelationEntityDto(
                    "pcs_rsc_order_set_1",
                    ResourceRelationType.ORDER_SET,
                    members=["d1", "d2", "d3", "d4", "d5", "d6"],
                    metadata={
                        "id": "pcs_rsc_order_set_1",
                        "sets": [
                            {
                                "id": "pcs_rsc_set_1",
                                "metadata": {
                                    "id": "pcs_rsc_set_1",
                                    "sequential": "true",
                                    "require-all": "true",
                                    "action": "start",
                                },
                                "members": ["d1", "d3", "d2"],
                            },
                            {
                                "id": "pcs_rsc_set_2",
                                "metadata": {
                                    "id": "pcs_rsc_set_2",
                                    "sequential": "false",
                                    "require-all": "false",
                                    "action": "stop",
                                },
                                "members": ["d6", "d5", "d4"],
                            },
                        ],
                        "kind": "Serialize",
                        "symmetrical": "true",
                    },
                ),
            },
        )
        for res in res_list:
            with self.subTest(resource=res):
                self.assertEqual(expected, obj.get_relations(res))

    def test_group(self):
        obj = lib.ResourceRelationsFetcher(
            fixture_cib(
                """
            <group id="g1">
              <primitive id="d1" class="c" provider="pcmk" type="Dummy"/>
              <primitive id="d2" class="c" provider="pcmk" type="Dummy"/>
            </group>
            """,
                "",
            )
        )
        expected = (
            {
                "d1": RelationEntityDto(
                    "d1",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["outer:g1"],
                    fixture_dummy_metadata("d1"),
                ),
                "d2": RelationEntityDto(
                    "d2",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["outer:g1"],
                    fixture_dummy_metadata("d2"),
                ),
                "g1": RelationEntityDto(
                    "g1",
                    ResourceRelationType.RSC_GROUP,
                    ["inner:g1"],
                    {"id": "g1"},
                ),
            },
            {
                "inner:g1": RelationEntityDto(
                    "inner:g1",
                    ResourceRelationType.INNER_RESOURCES,
                    ["d1", "d2"],
                    {"id": "g1"},
                ),
                "outer:g1": RelationEntityDto(
                    "outer:g1",
                    ResourceRelationType.OUTER_RESOURCE,
                    ["g1"],
                    {"id": "g1"},
                ),
            },
        )
        for res in ("d1", "d2", "g1"):
            with self.subTest(resource=res):
                self.assertEqual(expected, obj.get_relations(res))

    def _test_wrapper(self, wrapper_tag, rel_type):
        obj = lib.ResourceRelationsFetcher(
            fixture_cib(
                f"""
            <{wrapper_tag} id="w1">
              <primitive id="d1" class="c" provider="pcmk" type="Dummy"/>
            </{wrapper_tag}>
            """,
                "",
            )
        )
        expected = (
            {
                "d1": RelationEntityDto(
                    "d1",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["outer:w1"],
                    fixture_dummy_metadata("d1"),
                ),
                "w1": RelationEntityDto(
                    "w1", rel_type, ["inner:w1"], {"id": "w1"}
                ),
            },
            {
                "inner:w1": RelationEntityDto(
                    "inner:w1",
                    ResourceRelationType.INNER_RESOURCES,
                    ["d1"],
                    {"id": "w1"},
                ),
                "outer:w1": RelationEntityDto(
                    "outer:w1",
                    ResourceRelationType.OUTER_RESOURCE,
                    ["w1"],
                    {"id": "w1"},
                ),
            },
        )
        for res in ("d1", "w1"):
            with self.subTest(resource=res):
                self.assertEqual(expected, obj.get_relations(res))

    def test_clone(self):
        self._test_wrapper("clone", ResourceRelationType.RSC_CLONE)

    def test_master(self):
        self._test_wrapper("master", ResourceRelationType.RSC_CLONE)

    def test_bundle(self):
        self._test_wrapper("bundle", ResourceRelationType.RSC_BUNDLE)

    def test_cloned_group(self):
        obj = lib.ResourceRelationsFetcher(
            fixture_cib(
                """
            <clone id="c1">
                <group id="g1">
                  <primitive id="d1" class="c" provider="pcmk" type="Dummy"/>
                  <primitive id="d2" class="c" provider="pcmk" type="Dummy"/>
                </group>
            </clone>
            """,
                "",
            )
        )
        expected = (
            {
                "d1": RelationEntityDto(
                    "d1",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["outer:g1"],
                    fixture_dummy_metadata("d1"),
                ),
                "d2": RelationEntityDto(
                    "d2",
                    ResourceRelationType.RSC_PRIMITIVE,
                    ["outer:g1"],
                    fixture_dummy_metadata("d2"),
                ),
                "g1": RelationEntityDto(
                    "g1",
                    ResourceRelationType.RSC_GROUP,
                    ["inner:g1", "outer:c1"],
                    {"id": "g1"},
                ),
                "c1": RelationEntityDto(
                    "c1",
                    ResourceRelationType.RSC_CLONE,
                    ["inner:c1"],
                    {"id": "c1"},
                ),
            },
            {
                "inner:g1": RelationEntityDto(
                    "inner:g1",
                    ResourceRelationType.INNER_RESOURCES,
                    ["d1", "d2"],
                    {"id": "g1"},
                ),
                "outer:g1": RelationEntityDto(
                    "outer:g1",
                    ResourceRelationType.OUTER_RESOURCE,
                    ["g1"],
                    {"id": "g1"},
                ),
                "inner:c1": RelationEntityDto(
                    "inner:c1",
                    ResourceRelationType.INNER_RESOURCES,
                    ["g1"],
                    {"id": "c1"},
                ),
                "outer:c1": RelationEntityDto(
                    "outer:c1",
                    ResourceRelationType.OUTER_RESOURCE,
                    ["c1"],
                    {"id": "c1"},
                ),
            },
        )
        for res in ("d1", "d2", "g1", "c1"):
            with self.subTest(resource=res):
                self.assertEqual(expected, obj.get_relations(res))


class ResourceRelationTreeBuilder(TestCase):
    @staticmethod
    def primitive_fixture(_id, members):
        return RelationEntityDto(
            _id,
            ResourceRelationType.RSC_PRIMITIVE,
            members,
            fixture_dummy_metadata(_id),
        )

    def test_resource_not_present(self):
        with self.assertRaises(AssertionError):
            lib.ResourceRelationTreeBuilder({}, {}).get_tree("not_existing")

    def test_simple_order(self):
        resources_members = ["order-d1-d2-mandatory"]
        resources = {
            "d1": self.primitive_fixture("d1", resources_members),
            "d2": self.primitive_fixture("d2", resources_members),
        }
        relations = {
            "order-d1-d2-mandatory": RelationEntityDto(
                "order-d1-d2-mandatory",
                ResourceRelationType.ORDER,
                members=["d1", "d2"],
                metadata={
                    "id": "order-d1-d2-mandatory",
                    "first": "d1",
                    "first-action": "start",
                    "then": "d2",
                    "then-action": "start",
                    "kind": "Mandatory",
                },
            ),
        }
        expected = dict(
            relation_entity=dto.to_dict(resources["d2"]),
            is_leaf=False,
            members=[
                dict(
                    relation_entity=dto.to_dict(
                        relations["order-d1-d2-mandatory"]
                    ),
                    is_leaf=False,
                    members=[
                        dict(
                            relation_entity=dto.to_dict(resources["d1"]),
                            is_leaf=False,
                            members=[],
                        )
                    ],
                )
            ],
        )
        self.assertEqual(
            expected,
            dto.to_dict(
                lib.ResourceRelationTreeBuilder(resources, relations)
                .get_tree("d2")
                .to_dto()
            ),
        )

    def test_simple_order_set(self):
        res_list = ("d1", "d2", "d3", "d4", "d5", "d6")
        resources_members = ["pcs_rsc_order_set_1"]
        resources = {
            _id: self.primitive_fixture(_id, resources_members)
            for _id in res_list
        }
        relations = {
            "pcs_rsc_order_set_1": RelationEntityDto(
                "pcs_rsc_order_set_1",
                ResourceRelationType.ORDER_SET,
                members=["d1", "d2", "d3", "d4", "d5", "d6"],
                metadata={
                    "id": "pcs_rsc_order_set_1",
                    "sets": [
                        {
                            "id": "pcs_rsc_set_1",
                            "metadata": {
                                "id": "pcs_rsc_set_1",
                                "sequential": "true",
                                "require-all": "true",
                                "action": "start",
                            },
                            "members": ["d1", "d3", "d2"],
                        },
                        {
                            "id": "pcs_rsc_set_2",
                            "metadata": {
                                "id": "pcs_rsc_set_2",
                                "sequential": "false",
                                "require-all": "false",
                                "action": "stop",
                            },
                            "members": ["d6", "d5", "d4"],
                        },
                    ],
                    "kind": "Serialize",
                    "symmetrical": "true",
                },
            ),
        }
        get_res = lambda _id: dict(
            relation_entity=dto.to_dict(resources[_id]),
            is_leaf=False,
            members=[],
        )
        expected = dict(
            relation_entity=dto.to_dict(resources["d5"]),
            is_leaf=False,
            members=[
                dict(
                    relation_entity=dto.to_dict(
                        relations["pcs_rsc_order_set_1"]
                    ),
                    is_leaf=False,
                    members=[
                        get_res(_id) for _id in ("d1", "d2", "d3", "d4", "d6")
                    ],
                )
            ],
        )
        self.assertEqual(
            expected,
            dto.to_dict(
                lib.ResourceRelationTreeBuilder(resources, relations)
                .get_tree("d5")
                .to_dto()
            ),
        )

    def test_simple_in_group(self):
        resources_members = ["outer:g1"]
        resources = {
            "d1": self.primitive_fixture("d1", resources_members),
            "d2": self.primitive_fixture("d2", resources_members),
            "g1": RelationEntityDto("g1", "group", ["inner:g1"], {"id": "g1"}),
        }
        relations = {
            "inner:g1": RelationEntityDto(
                "inner:g1",
                ResourceRelationType.INNER_RESOURCES,
                ["d1", "d2"],
                {"id": "g1"},
            ),
            "outer:g1": RelationEntityDto(
                "outer:g1",
                ResourceRelationType.OUTER_RESOURCE,
                ["g1"],
                {"id": "g1"},
            ),
        }
        expected = dict(
            relation_entity=dto.to_dict(resources["d1"]),
            is_leaf=False,
            members=[
                dict(
                    relation_entity=dto.to_dict(relations["outer:g1"]),
                    is_leaf=False,
                    members=[
                        dict(
                            relation_entity=dto.to_dict(resources["g1"]),
                            is_leaf=False,
                            members=[
                                dict(
                                    relation_entity=dto.to_dict(
                                        relations["inner:g1"]
                                    ),
                                    is_leaf=False,
                                    members=[
                                        dict(
                                            relation_entity=dto.to_dict(
                                                resources["d2"]
                                            ),
                                            is_leaf=False,
                                            members=[],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )
        self.assertEqual(
            expected,
            dto.to_dict(
                lib.ResourceRelationTreeBuilder(resources, relations)
                .get_tree("d1")
                .to_dto()
            ),
        )

    def test_order_loop(self):
        resources_members = ["order-d1-d2-mandatory", "order-d2-d1-mandatory"]
        resources = {
            "d1": self.primitive_fixture("d1", resources_members),
            "d2": self.primitive_fixture("d2", resources_members),
        }
        order_fixture = lambda r1, r2: RelationEntityDto(
            f"order-{r1}-{r2}-mandatory",
            ResourceRelationType.ORDER,
            members=[r1, r2],
            metadata={
                "id": f"order-{r1}-{r2}-mandatory",
                "first": r1,
                "first-action": "start",
                "then": r2,
                "then-action": "start",
                "kind": "Mandatory",
            },
        )
        relations = {
            "order-d1-d2-mandatory": order_fixture("d1", "d2"),
            "order-d2-d1-mandatory": order_fixture("d2", "d1"),
        }
        expected = dict(
            relation_entity=dto.to_dict(resources["d1"]),
            is_leaf=False,
            members=[
                dict(
                    relation_entity=dto.to_dict(
                        relations["order-d1-d2-mandatory"]
                    ),
                    is_leaf=False,
                    members=[
                        dict(
                            relation_entity=dto.to_dict(resources["d2"]),
                            is_leaf=False,
                            members=[
                                dict(
                                    relation_entity=dto.to_dict(
                                        relations["order-d2-d1-mandatory"]
                                    ),
                                    is_leaf=True,
                                    members=[],
                                ),
                            ],
                        ),
                    ],
                ),
                dict(
                    relation_entity=dto.to_dict(
                        relations["order-d2-d1-mandatory"]
                    ),
                    is_leaf=False,
                    members=[
                        dict(
                            relation_entity=dto.to_dict(resources["d2"]),
                            is_leaf=True,
                            members=[],
                        ),
                    ],
                ),
            ],
        )
        self.assertEqual(
            expected,
            dto.to_dict(
                lib.ResourceRelationTreeBuilder(resources, relations)
                .get_tree("d1")
                .to_dto()
            ),
        )
